from rattr.analyser.context import Name, SymbolTable


class TestSymbolTable:
    def test_add(self):
        symbol_table = SymbolTable()

        symbols = [Name("a"), Name("b")]

        for symbol in symbols:
            symbol_table.add(symbol)

        assert list(symbol_table.names()) == [s.name for s in symbols]
        assert list(symbol_table.symbols()) == symbols

    def test_remove(self):
        # Remove when absent
        symbol_table = SymbolTable()
        symbol_table.remove("a")
        assert list(symbol_table.names()) == list()

        # Remove when present
        symbol_table = SymbolTable()

        symbol_table.add(Name("a"))
        symbol_table.add(Name("b"))
        assert list(symbol_table.names()) == ["a", "b"]

        symbol_table.remove("b")
        assert list(symbol_table.names()) == ["a"]

    def test_get(self):
        symbol_table = SymbolTable()

        symbol = Name("a")
        symbol_table.add(symbol)

        assert symbol_table.get("a") is symbol



"""Rattr functions for producing and displaying results."""
import json

from rattr.analyser.context import Symbol
from rattr.analyser.ir_dag import IrDagNode
from rattr.analyser.types import FileIR, FileResults, ImportsIR


class ResultsEncoder(json.JSONEncoder):
    """Return the results encoded as JSON."""

    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)

        if isinstance(obj, Symbol):
            return repr(obj)

        return super().default(obj)


def generate_results_from_ir(file_ir: FileIR, imports_ir: ImportsIR) -> FileResults:
    """Generate the final results from the given IR.

    Removing cycles:

        Direct recursion:
            A -> A

            R(A)    = r(A) ∪ r(A)                               (by definition)
            .       = r(A)                                      (idempotent)

            where R(A) is the final result of a A;
            .     r(A) is the partial result of A

        Indirect recursion:
            A -> B -> ... -> C -> A

            R(A)    =  r(A)         ∪ r(B) ∪ ... ∪ r(C) ∪ r(A)  (by definition)
            .       = (r(A) ∪ r(A)) ∪ r(B) ∪ ... ∪ r(C)         (commutativity)
            .       =  r(A)         ∪ r(B) ∪ ... ∪ r(C)         (idempotent)

            where R(N) is the final result of a N, for any node N;
            .     r(N) is the partial result of N, for any node N

        I.e. cycles in the DAG of function calls from from root function can be
        simplified by "ignoring" previously visited functions, given the
        arguments are the same (as visiting them a second time will not affect
        the final result).

    """
    simplified: FileResults = dict()

    for foc, foc_ir in file_ir.items():
        ir_dag = IrDagNode(None, foc, foc_ir, file_ir, imports_ir)
        ir_dag.populate()

        composed_ir = ir_dag.simplify()

        simplified[foc.name] = {
            "sets": {s.name for s in composed_ir["sets"]},
            "gets": {s.name for s in composed_ir["gets"]},
            "dels": {s.name for s in composed_ir["dels"]},
            "calls": {s.name for s in composed_ir["calls"]},
        }

    return simplified



"""Tests for module/file level features."""
from unittest import mock

import pytest

from rattr.analyser.context import RootContext
from rattr.analyser.context.symbol import Call, Func, Name
from rattr.analyser.file import FileAnalyser


class TestModuleLevelStatements:
    def test_assignment(self, parse):
        _ast = parse(
            """
            one = 1
            one = 1
            two = 2
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        assert results == {}

    def test_typed_assignment(self, parse):
        _ast = parse(
            """
            one: int = 1
            one: int = 1
            two: int = 2
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        assert results == {}

    def test_augmented_assignment(self, parse):
        _ast = parse(
            """
            one += 1
            one += 1
            two += 2
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        assert results == {}

    @pytest.mark.py_3_8_plus()
    def test_walrus_operator(self, parse):
        _ast = parse(
            """
            x = (y := 1)
            x = (y := 2)
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        assert results == {}

    def test_multiple_assignment(self, parse):
        _ast = parse(
            """
            a, b = some_tuple
            x, y = 1, 2
            z = 3, 4
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        assert results == {}

    def test_lambda(self, parse):
        # Anonymous
        _ast = parse(
            """
            lambda *a, **k: a.attr
            """
        )
        with mock.patch("sys.exit") as _exit:
            FileAnalyser(_ast, RootContext(_ast)).analyse()

        # Named
        _ast = parse(
            """
            name = lambda *a, **k: a.attr
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("name", [], "a", "k"): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("a.attr", "a"),
                },
                "sets": set(),
            },
        }

        assert results == expected

    @pytest.mark.py_3_8_plus()
    def test_walrus_multiple_assignment(self, parse):
        # Walrus multi assign w/ lambda
        _ast = parse(
            """
            other = (alpha, beta := 1)
            other, another = (alpha, beta := 1)
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {}

        assert results == expected

    @pytest.mark.py_3_8_plus()
    def test_walrus_multiple_assignment_list(self, parse):
        # Walrus multi assign w/ lambda
        _ast = parse(
            """
            other = [alpha, beta := 1]
            other, another = [alpha, beta := 1]
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {}

        assert results == expected

    @pytest.mark.py_3_8_plus()
    def test_walrus_lambda(self, parse):
        # Walrus'd Lambda
        _ast = parse(
            """
            other = (name := lambda *a, **k: a.attr)
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("name", [], "a", "k"): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("a.attr", "a"),
                },
                "sets": set(),
            },
            Func("other", [], "a", "k"): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("a.attr", "a"),
                },
                "sets": set(),
            },
        }

        assert results == expected

    @pytest.mark.py_3_8_plus()
    def test_walrus_multiple_assignment_lambda(self, parse):
        # Walrus multi assign w/ lambda
        _ast = parse(
            """
            other = (alpha, beta := lambda *a, **k: a.attr)
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("beta", [], "a", "k"): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("a.attr", "a"),
                },
                "sets": set(),
            },
        }

        assert results == expected

    @pytest.mark.py_3_8_plus()
    def test_walrus_multiple_assignment_lambda_list(self, parse):
        # Walrus multi assign w/ lambda
        _ast = parse(
            """
            other = [alpha, beta := lambda *a, **k: a.attr]
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("beta", [], "a", "k"): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("a.attr", "a"),
                },
                "sets": set(),
            },
        }

        assert results == expected



from unittest import mock

from rattr.analyser.context import Call, Func, Import, Name
from rattr.analyser.context.symbol import Class
from rattr.analyser.results import generate_results_from_ir


class TestResults:
    def test_generate_results_from_ir_no_calls(self):
        # No calls
        fn_ir = {
            "sets": {
                Name("arg.attr", "arg"),
            },
            "gets": {
                Name("arg.another_attr", "arg"),
            },
            "dels": set(),
            "calls": set(),
        }
        file_ir = {Func("fn", ["arg"], None, None): fn_ir}

        expected = {
            "fn": {
                "sets": {"arg.attr"},
                "gets": {"arg.another_attr"},
                "dels": set(),
                "calls": set(),
            }
        }

        assert generate_results_from_ir(file_ir, dict()) == expected

    def test_generate_results_from_ir_simple(self):
        # Calls
        fn_a = Func("fn_a", ["arg"], None, None)
        fn_b = Func("fn_b", ["arg_b"], None, None)
        fn_a_ir = {
            "sets": {
                Name("arg.attr", "arg"),
            },
            "gets": {
                Name("arg.another_attr", "arg"),
            },
            "dels": set(),
            "calls": {
                Call("fn_b()", ["arg"], {}, target=fn_b),
            },
        }
        fn_b_ir = {
            "sets": {
                Name("arg_b.set_in_fn_b", "arg_b"),
            },
            "gets": {
                Name("arg_b.get_in_fn_b", "arg_b"),
            },
            "dels": set(),
            "calls": set(),
        }
        file_ir = {
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
        }

        expected = {
            "fn_a": {
                "sets": {"arg.attr", "arg.set_in_fn_b"},
                "gets": {"arg.another_attr", "arg.get_in_fn_b"},
                "dels": set(),
                "calls": {"fn_b()"},
            },
            "fn_b": {
                "sets": {"arg_b.set_in_fn_b"},
                "gets": {"arg_b.get_in_fn_b"},
                "dels": set(),
                "calls": set(),
            },
        }

        assert generate_results_from_ir(file_ir, dict()) == expected

    def test_generate_results_from_ir_direct_recursion(self):
        # Direct recursion
        fn_ir = {
            "sets": {
                Name("arg.attr", "arg"),
            },
            "gets": {
                Name("arg.another_attr", "arg"),
            },
            "dels": set(),
            "calls": {Call("fn()", ["arg"], {})},
        }
        file_ir = {Func("fn", ["arg"], None, None): fn_ir}

        expected = {
            "fn": {
                "sets": {"arg.attr"},
                "gets": {"arg.another_attr"},
                "dels": set(),
                "calls": {"fn()"},
            }
        }

        assert generate_results_from_ir(file_ir, dict()) == expected

    def test_generate_results_from_ir_indirect_recursion(self):
        # Indirect recursion
        fn_a = Func("fn_a", ["arg_a"], None, None)
        fn_b = Func("fn_b", ["arg_b"], None, None)
        fn_a_ir = {
            "sets": {Name("arg_a.get_from_a", "arg_a")},
            "gets": set(),
            "dels": set(),
            "calls": {Call("fn_b()", ["arg_a"], {}, target=fn_b)},
        }
        fn_b_ir = {
            "sets": {Name("arg_b.get_from_b", "arg_b")},
            "gets": set(),
            "dels": set(),
            "calls": {Call("fn_a()", ["arg_b"], {}, target=fn_a)},
        }
        file_ir = {
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
        }

        expected = {
            "fn_a": {
                "sets": {"arg_a.get_from_a", "arg_a.get_from_b"},
                "gets": set(),
                "dels": set(),
                "calls": {"fn_b()"},
            },
            "fn_b": {
                "sets": {"arg_b.get_from_a", "arg_b.get_from_b"},
                "gets": set(),
                "dels": set(),
                "calls": {"fn_a()"},
            },
        }

        assert generate_results_from_ir(file_ir, dict()) == expected

    def test_generate_results_from_ir_child_has_direct_recursion(self):
        fn_a = Func("fn_a", ["x"], None, None)
        fn_b = Func("fn_b", ["arg"], None, None)
        fn_a_ir = {
            "sets": set(),
            "gets": {Name("x.attr", "x")},
            "dels": set(),
            "calls": {Call("fn_b()", ["x"], {}, target=fn_b)},
        }
        fn_b_ir = {
            "sets": set(),
            "gets": {Name("arg.field", "arg")},
            "dels": set(),
            "calls": {Call("fn_b()", ["arg"], {}, target=fn_b)},
        }
        file_ir = {
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
        }

        expected = {
            "fn_a": {
                "sets": set(),
                "gets": {"x.attr", "x.field"},
                "dels": set(),
                "calls": {"fn_b()"},
            },
            "fn_b": {
                "sets": set(),
                "gets": {"arg.field"},
                "dels": set(),
                "calls": {"fn_b()"},
            },
        }

        assert generate_results_from_ir(file_ir, dict()) == expected

    def test_generate_results_from_ir_child_has_indirect_recursion(self):
        fn_a = Func("fn_a", ["a"], None, None)
        fn_b = Func("fn_b", ["b"], None, None)
        fn_c = Func("fn_c", ["c"], None, None)
        fn_a_ir = {
            "sets": set(),
            "gets": {Name("a.in_a", "a")},
            "dels": set(),
            "calls": {Call("fn_b()", ["a"], {}, target=fn_b)},
        }
        fn_b_ir = {
            "sets": set(),
            "gets": {Name("b.in_b", "b")},
            "dels": set(),
            "calls": {Call("fn_c()", ["b"], {}, target=fn_c)},
        }
        fn_c_ir = {
            "sets": set(),
            "gets": {Name("c.in_c", "c")},
            "dels": set(),
            "calls": {Call("fn_b()", [], {"b": "c"}, target=fn_b)},
        }
        file_ir = {
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
            fn_c: fn_c_ir,
        }

        expected = {
            "fn_a": {
                "sets": set(),
                "gets": {"a.in_a", "a.in_b", "a.in_c"},
                "dels": set(),
                "calls": {"fn_b()"},
            },
            "fn_b": {
                "sets": set(),
                "gets": {"b.in_b", "b.in_c"},
                "dels": set(),
                "calls": {"fn_c()"},
            },
            "fn_c": {
                "sets": set(),
                "gets": {"c.in_b", "c.in_c"},
                "dels": set(),
                "calls": {"fn_b()"},
            },
        }

        assert generate_results_from_ir(file_ir, dict()) == expected

    def test_generate_results_from_ir_repeated_calls(self):
        # Repeated calls that should be ignored
        fn_a = Func("fn_a", ["a"], None, None)
        fn_b = Func("fn_b", ["b"], None, None)
        fn_a_ir = {
            "sets": set(),
            "gets": {Name("a.in_a", "a")},
            "dels": set(),
            "calls": {
                Call("fn_b()", ["a.attr"], {}, target=fn_b),
                Call("fn_b()", ["a.attr"], {}, target=fn_b),
            },
        }
        fn_b_ir = {
            "sets": set(),
            "gets": {Name("b.in_b", "b")},
            "dels": set(),
            "calls": set(),
        }
        file_ir = {
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
        }

        expected = {
            "fn_a": {
                "sets": set(),
                "gets": {"a.in_a", "a.attr.in_b"},
                "dels": set(),
                "calls": {"fn_b()"},
            },
            "fn_b": {
                "sets": set(),
                "gets": {"b.in_b"},
                "dels": set(),
                "calls": set(),
            },
        }

        assert generate_results_from_ir(file_ir, dict()) == expected

        # Repeated calls that should not be ignored
        fn_a = Func("fn_a", ["a"], None, None)
        fn_b = Func("fn_b", ["b"], None, None)
        fn_a_ir = {
            "sets": set(),
            "gets": {Name("a.in_a", "a")},
            "dels": set(),
            "calls": {
                Call("fn_b()", ["a.attr_one"], {}, target=fn_b),
                Call("fn_b()", ["a.attr_two"], {}, target=fn_b),
            },
        }
        fn_b_ir = {
            "sets": set(),
            "gets": {Name("b.in_b", "b")},
            "dels": set(),
            "calls": set(),
        }
        file_ir = {
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
        }

        expected = {
            "fn_a": {
                "sets": set(),
                "gets": {"a.in_a", "a.attr_one.in_b", "a.attr_two.in_b"},
                "dels": set(),
                "calls": {"fn_b()"},
            },
            "fn_b": {
                "sets": set(),
                "gets": {"b.in_b"},
                "dels": set(),
                "calls": set(),
            },
        }

        assert generate_results_from_ir(file_ir, dict()) == expected

    def test_imports_ir(self, file_ir_from_dict):
        # Simple
        imports_ir = {
            "module": file_ir_from_dict(
                {
                    Func("act", ["arg"], None, None): {
                        "sets": set(),
                        "gets": {
                            Name("arg.attr", "arg"),
                        },
                        "dels": set(),
                        "calls": set(),
                    }
                }
            )
        }

        _i = Import("act", "module.act")
        _i.module_name = "module"
        _i.module_spec = mock.Mock()

        fn = Func("fn", ["ms"], None, None)
        fn_ir = {
            "sets": set(),
            "gets": set(),
            "dels": set(),
            "calls": {Call("act()", ["ms"], {}, target=_i)},
        }
        file_ir = {
            fn: fn_ir,
        }

        expected = {
            "fn": {
                "sets": set(),
                "gets": {"ms.attr"},
                "dels": set(),
                "calls": {"act()"},
            }
        }

        assert generate_results_from_ir(file_ir, imports_ir) == expected

        # Chained
        _i_second = Import("second", "chained.second")
        _i_second.module_name = "chained"
        _i_second.module_spec = mock.Mock()

        imports_ir = {
            "module": file_ir_from_dict(
                {
                    Func("first", ["arrg"], None, None): {
                        "sets": set(),
                        "gets": set(),
                        "dels": set(),
                        "calls": {Call("second", ["arrg"], {}, target=_i_second)},
                    }
                }
            ),
            "chained": file_ir_from_dict(
                {
                    Func("second", ["blarg"], None, None): {
                        "sets": set(),
                        "gets": {
                            Name("blarg._attr", "blarg"),
                        },
                        "dels": set(),
                        "calls": set(),
                    }
                }
            ),
        }

        _i_module = Import("first", "module.first")
        _i_module.module_name = "module"
        _i_module.module_spec = mock.Mock()

        fn = Func("fn", ["flarg"], None, None)
        fn_ir = {
            "sets": set(),
            "gets": set(),
            "dels": set(),
            "calls": {Call("first()", ["flarg"], {}, target=_i_module)},
        }
        file_ir = {
            fn: fn_ir,
        }

        expected = {
            "fn": {
                "sets": set(),
                "gets": {"flarg._attr"},
                "dels": set(),
                "calls": {
                    "first()",
                },
            }
        }

        assert generate_results_from_ir(file_ir, imports_ir) == expected

    def test_class(self):
        cls_inst = Class("SomeClass", ["self", "arg"], None, None)
        cls_inst_ir = {
            "sets": {
                Name("self.my_attr", "self"),
            },
            "gets": {
                Name("arg.attr_in_init", "arg"),
            },
            "dels": set(),
            "calls": set(),
        }
        cls_inst_sm = Func("SomeClass.static", ["flarg"], None, None)
        cls_inst_sm_ir = {
            "sets": {
                Name("flarg.attr_in_static", "flarg"),
            },
            "gets": set(),
            "dels": set(),
            "calls": set(),
        }
        fn = Func("i_call_them", ["marg"], None, None)
        fn_ir = {
            "sets": {
                Name("instance"),
            },
            "gets": set(),
            "dels": set(),
            "calls": {
                Call("SomeClass()", ["instance", "marg"], {}, target=cls_inst),
                Call("SomeClass.static()", ["marg"], {}, target=cls_inst_sm),
            },
        }
        file_ir = {
            cls_inst: cls_inst_ir,
            cls_inst_sm: cls_inst_sm_ir,
            fn: fn_ir,
        }

        expected = {
            "SomeClass": {
                "sets": {"self.my_attr"},
                "gets": {"arg.attr_in_init"},
                "dels": set(),
                "calls": set(),
            },
            "SomeClass.static": {
                "sets": {"flarg.attr_in_static"},
                "gets": set(),
                "dels": set(),
                "calls": set(),
            },
            "i_call_them": {
                "sets": {
                    "instance",
                    "instance.my_attr",
                    "marg.attr_in_static",
                },
                "gets": {
                    "marg.attr_in_init",
                },
                "dels": set(),
                "calls": {
                    "SomeClass()",
                    "SomeClass.static()",
                },
            },
        }

        assert generate_results_from_ir(file_ir, dict()) == expected



"""Annotations to be imported for use by Rattr analysed code."""

from typing import Dict, List, Optional, Set, Tuple

# A set of names, e.g.: "@BinOp", "x.attr", "var_one"
Names = Set[str]

# LHS: positional args, RHS: named args
CalleeArgs = Tuple[List[str], Dict[str, str]]

# LHS: callee name, RHS: callee args
Calls = List[Tuple[str, CalleeArgs]]


def rattr_ignore():
    """Do not parse the decorated function, thus omitted from the results."""

    def _inner(f, *args, **kwargs):
        return f

    return _inner


def rattr_results(
    gets: Optional[Names] = None,
    sets: Optional[Names] = None,
    dels: Optional[Names] = None,
    calls: Optional[Calls] = None,
):
    """Explicitly provide results for a function which will then be ignored."""

    def _inner(f, *args, **kwargs):
        return f

    return _inner



import ast
import sys
from contextlib import contextmanager
from os.path import dirname, join
from typing import Iterable

import pytest

import rattr
from rattr.analyser.base import CustomFunctionAnalyser, CustomFunctionHandler
from rattr.analyser.context import Context, RootContext
from rattr.analyser.context.symbol import Call, Name
from rattr.analyser.types import FileIR, FuncOrAsyncFunc, FunctionIR
from rattr.analyser.util import LOCAL_VALUE_PREFIX
from rattr.cli.parser import (
    Cache,
    ExcludeImports,
    ExcludePatterns,
    FollowImports,
    Output,
    ShowPath,
    ShowWarnings,
    StrictOrPermissive,
)


def pytest_configure(config):
    config.addinivalue_line("markers", "pypy: mark test to run only under pypy")
    config.addinivalue_line(
        "markers", "py_3_8_plus: mark test to run only under Python 3.8+"
    )


def pytest_collection_modifyitems(config, items):
    """Alter the collected tests."""
    filter_marked(items, "pypy", is_pypy())
    filter_marked(items, "py_3_8_plus", is_python_3_8_plus())


def filter_marked(items: Iterable, mark: str, condition: bool):
    """Remove the marked items if the condition is `False`."""
    marked = list()
    unmarked = list()

    for item in items:
        if item.get_closest_marker(mark):
            marked.append(item)
        else:
            unmarked.append(item)

    if condition:
        items[:] = marked + unmarked
    else:
        items[:] = unmarked


def is_pypy():
    """Return `True` if running under pypy."""
    try:
        import __pypy__

        return True
    except ModuleNotFoundError:
        return False


def is_python_3_8_plus():
    """Return `True` if running under Python 3.8 plus."""
    if sys.version_info.major != 3:
        raise NotImplementedError

    if sys.version_info.minor >= 8:
        return True
    else:
        return False


@pytest.fixture
def parse():
    def _inner(source: str) -> ast.AST:
        """Return the parsed AST for the given code, use relative indentation.

        Assume usage is:
            parse('''
                <source code>       # first line sets the base indent
                <source code>
                    <source code>   # this is indented once (relative)
                <source code>
            ''')

        Require the opening and closing blank lines.
        """
        lines = list()
        source_lines = source.splitlines()[1:-1]

        if not len(source_lines):
            raise ValueError("Incorrect source code formatting")

        # Determine whitespace from first line
        indent = str()
        for c in source_lines[0]:
            if not c.isspace():
                break
            indent += c

        # Strip whitespace from all lines
        for line in source_lines:
            if line.startswith(indent):
                lines.append(line[len(indent) :])
            elif line == "":
                lines.append(line)
            else:
                raise SyntaxError("Incorrect indentation in test code")

        return ast.parse("\n".join(lines))

    return _inner


@pytest.fixture
def RootSymbolTable():
    def _inner(*args):
        """Create a root context with the addition of the **kwargs."""
        symtab = RootContext(ast.Module(body=[])).symbol_table

        for s in args:
            symtab.add(s)

        return symtab

    return _inner


@pytest.fixture
def config():
    @contextmanager
    def _inner(attr, value):
        if not hasattr(rattr.config, attr):
            raise AttributeError

        _previous = getattr(rattr.config, attr)
        setattr(rattr.config, attr, value)
        yield
        setattr(rattr.config, attr, _previous)

    return _inner


@pytest.fixture
def stdlib_modules():
    # Scraped from python.org
    scraped = {
        "string",
        "re",
        "difflib",
        "textwrap",
        "unicodedata",
        "stringprep",
        "rlcompleter",
        "struct",
        "codecs",
        "datetime",
        "calendar",
        "collections",
        "abc",
        "heapq",
        "bisect",
        "array",
        "weakref",
        "types",
        "copy",
        "pprint",
        "reprlib",
        "enum",
        "numbers",
        "math",
        "cmath",
        "decimal",
        "fractions",
        "random",
        "statistics",
        "itertools",
        "functools",
        "operator",
        "pathlib",
        "fileinput",
        "stat",
        "filecmp",
        "tempfile",
        "glob",
        "fnmatch",
        "linecache",
        "shutil",
        "pickle",
        "copyreg",
        "shelve",
        "dbm",
        "sqlite3",
        "zlib",
        "gzip",
        "bz2",
        "lzma",
        "zipfile",
        "tarfile",
        "csv",
        "configparser",
        "netrc",
        "xdrlib",
        "plistlib",
        "hashlib",
        "hmac",
        "secrets",
        "os",
        "io",
        "time",
        "argparse",
        "getopt",
        "logging",
        "getpass",
        "curses",
        "platform",
        "errno",
        "ctypes",
        "threading",
        "multiprocessing",
        "concurrent",
        "subprocess",
        "sched",
        "queue",
        "_thread",
        "asyncio",
        "asyncio",
        "socket",
        "ssl",
        "select",
        "selectors",
        "asyncore",
        "asynchat",
        "signal",
        "mmap",
        "email",
        "json",
        "mailcap",
        "mailbox",
        "mimetypes",
        "base64",
        "binhex",
        "binascii",
        "quopri",
        "uu",
        "html",
        "webbrowser",
        "cgi",
        "cgitb",
        "wsgiref",
        "urllib",
        "http",
        "ftplib",
        "poplib",
        "imaplib",
        "nntplib",
        "smtplib",
        "smtpd",
        "telnetlib",
        "uuid",
        "socketserver",
        "xmlrpc",
        "ipaddress",
        "audioop",
        "aifc",
        "sunau",
        "wave",
        "chunk",
        "colorsys",
        "imghdr",
        "sndhdr",
        "ossaudiodev",
        "gettext",
        "locale",
        "turtle",
        "cmd",
        "shlex",
        "typing",
        "pydoc",
        "doctest",
        "unittest",
        "test",
        "bdb",
        "faulthandler",
        "pdb",
        "timeit",
        "trace",
        "tracemalloc",
        "distutils",
        "venv",
        "zipapp",
        "sys",
        "sysconfig",
        "builtins",
        "warnings",
        "dataclasses",
        "contextlib",
        "abc",
        "atexit",
        "traceback",
        "gc",
        "inspect",
        "site",
        "code",
        "codeop",
        "pkgutil",
        "modulefinder",
        "runpy",
        "importlib",
        "ast",
        "symtable",
        "token",
        "keyword",
        "tokenize",
        "tabnanny",
        "pyclbr",
        "py_compile",
        "compileall",
        "dis",
        "pickletools",
        "spwd",
        "crypt",
        "tty",
        "pty",
        "pipes",
        "nis",
        "optparse",
        "imp",
    }

    return scraped


@pytest.fixture
def builtins():
    generated = {
        "abs",
        "all",
        "any",
        "ascii",
        "bin",
        "bool",
        "breakpoint",
        "bytearray",
        "bytes",
        "callable",
        "chr",
        "classmethod",
        "compile",
        "complex",
        "copyright",
        "credits",
        "delattr",
        "dict",
        "dir",
        "divmod",
        "enumerate",
        "eval",
        "exec",
        "filter",
        "float",
        "format",
        "frozenset",
        "getattr",
        "globals",
        "hasattr",
        "hash",
        "help",
        "hex",
        "id",
        "input",
        "int",
        "isinstance",
        "issubclass",
        "iter",
        "len",
        "license",
        "list",
        "locals",
        "map",
        "max",
        "memoryview",
        "min",
        "next",
        "object",
        "oct",
        "open",
        "ord",
        "pow",
        "print",
        "property",
        "range",
        "repr",
        "reversed",
        "round",
        "set",
        "setattr",
        "slice",
        "sorted",
        "staticmethod",
        "str",
        "sum",
        "super",
        "tuple",
        "type",
        "vars",
        "zip",
    }

    return generated


@pytest.fixture
def snippet():
    def _inner(relative_path: str):
        return join(dirname(__file__), "snippets", relative_path)

    return _inner


@pytest.fixture
def file_ir_from_dict():
    def _inner(ir):
        # Make quasi-context
        ctx = Context(None)
        ctx.add_all(ir.keys())

        # Create FileIR
        file_ir = FileIR(ctx)
        file_ir._file_ir = ir

        return file_ir

    return _inner


class _PrintBuiltinAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "print"

    @property
    def qualified_name(self) -> str:
        return "print"

    def on_def(self, name: str, node: FuncOrAsyncFunc, ctx: Context) -> FunctionIR:
        return {
            "sets": {
                Name("set_in_print_def"),
            },
            "gets": {
                Name("get_in_print_def"),
            },
            "dels": {
                Name("del_in_print_def"),
            },
            "calls": {
                Call("call_in_print_def", [], {}, None),
            },
        }

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIR:
        return {
            "sets": {
                Name("set_in_print"),
            },
            "gets": {
                Name("get_in_print"),
            },
            "dels": {
                Name("del_in_print"),
            },
            "calls": {
                Call("call_in_print", [], {}, None),
            },
        }


@pytest.fixture
def PrintBuiltinAnalyser():
    return _PrintBuiltinAnalyser


class _ExampleFuncAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "example"

    @property
    def qualified_name(self) -> str:
        return "module.example"

    def on_def(self, name: str, node: FuncOrAsyncFunc, ctx: Context) -> FunctionIR:
        return {
            "sets": {
                Name("set_in_example_def"),
            },
            "gets": {
                Name("get_in_example_def"),
            },
            "dels": {
                Name("del_in_example_def"),
            },
            "calls": {
                Call("call_in_example_def", [], {}, None),
            },
        }

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIR:
        return {
            "sets": {
                Name("set_in_example"),
            },
            "gets": {
                Name("get_in_example"),
            },
            "dels": {
                Name("del_in_example"),
            },
            "calls": {
                Call("call_in_example", [], {}, None),
            },
        }


@pytest.fixture
def ExampleFuncAnalyser():
    return _ExampleFuncAnalyser


@pytest.fixture
def handler():
    handler = CustomFunctionHandler([_PrintBuiltinAnalyser()], [_ExampleFuncAnalyser()])

    return handler


@pytest.fixture
def constant():
    """Return the version specific name for a constant of the given type.

    In Python <= 3.7, constants are named as such "@Str", "@Num", etc.

    In Python >= 3.8, constants are all named "@Constant".

    """

    def _inner(node_type: str):
        if sys.version_info.major != 3:
            raise AssertionError

        if sys.version_info.minor <= 7:
            return f"{LOCAL_VALUE_PREFIX}{node_type}"
        else:
            return f"{LOCAL_VALUE_PREFIX}Constant"

    return _inner


@pytest.fixture
def illegal_field_name():
    return "illegal-field"


@pytest.fixture
def toml_dict_with_illegal_fields():
    return {
        FollowImports.ARG_LONG_NAME: 3,
        ExcludeImports.ARG_LONG_NAME: [
            "a\\.b\\.c",
            "a\\.b.*",
            "a\\.b\\.c\\.e",
            "a\\.b\\.c.*",
        ],
        ExcludePatterns.ARG_LONG_NAME: ["a_.*", "b_.*", "_.*"],
        ShowWarnings.ARG_LONG_NAME: "all",
        illegal_field_name: "full",  # bad field name
        StrictOrPermissive.PERMISSIVE_ARG_LONG_NAME: 1,
        Output.SHOW_IR_ARG_LONG_NAME: True,
        Output.SHOW_RESULTS_ARG_LONG_NAME: True,
        Output.SHOW_STATS_ARG_LONG_NAME: True,
        Cache.ARG_LONG_NAME: "cache.json",
    }


@pytest.fixture
def toml_dict_with_bad_field_types_and_error1():
    return {
        FollowImports.ARG_LONG_NAME: "abcd",  # bad field type
        ExcludeImports.ARG_LONG_NAME: [
            "a\\.b\\.c",
            "a\\.b.*",
            "a\\.b\\.c\\.e",
            "a\\.b\\.c.*",
        ],
        ExcludePatterns.ARG_LONG_NAME: ["a_.*", "b_.*", "_.*"],
        ShowWarnings.ARG_LONG_NAME: "all",
        ShowPath.ARG_LONG_NAME: "full",
        StrictOrPermissive.PERMISSIVE_ARG_LONG_NAME: 1,
        Output.SHOW_IR_ARG_LONG_NAME: True,
        Output.SHOW_RESULTS_ARG_LONG_NAME: True,
        Output.SHOW_STATS_ARG_LONG_NAME: True,
        Cache.ARG_LONG_NAME: "rattr/cli/parser.py",
    }, (
        f"Error parsing pyproject.toml. Arg: '{FollowImports.ARG_LONG_NAME}' "
        f"is of wrong type: 'str'. Expected type: "
        f"'{FollowImports.TOML_ARG_NAME_ARG_TYPE_MAP[FollowImports.ARG_LONG_NAME].__name__}'."  # noqa: E501
    )


@pytest.fixture
def toml_dict_with_bad_field_types_and_error2():
    return {
        FollowImports.ARG_LONG_NAME: 3,
        ExcludeImports.ARG_LONG_NAME: [
            "a\\.b\\.c",
            "a\\.b.*",
            "a\\.b\\.c\\.e",
            "a\\.b\\.c.*",
        ],
        ExcludePatterns.ARG_LONG_NAME: ["a_.*", "b_.*", "_.*"],
        ShowWarnings.ARG_LONG_NAME: "all",
        ShowPath.ARG_LONG_NAME: "full",
        StrictOrPermissive.PERMISSIVE_ARG_LONG_NAME: 1,
        Output.SHOW_IR_ARG_LONG_NAME: 4,  # bad field type
        Output.SHOW_RESULTS_ARG_LONG_NAME: True,
        Output.SHOW_STATS_ARG_LONG_NAME: True,
        Cache.ARG_LONG_NAME: "rattr/cli/parser.py",
    }, (
        f"Error parsing pyproject.toml. Arg: '{Output.SHOW_IR_ARG_LONG_NAME}' "
        f"is of wrong type: 'int'. Expected type: "
        f"'{Output.TOML_ARG_NAME_ARG_TYPE_MAP[Output.SHOW_IR_ARG_LONG_NAME].__name__}'."
    )


@pytest.fixture
def toml_dict_with_bad_field_types_and_error3():
    return {
        FollowImports.ARG_LONG_NAME: 3,
        ExcludeImports.ARG_LONG_NAME: [
            "a\\.b\\.c",
            "a\\.b.*",
            "a\\.b\\.c\\.e",
            "a\\.b\\.c.*",
        ],
        ExcludePatterns.ARG_LONG_NAME: True,  # bad field type
        ShowWarnings.ARG_LONG_NAME: "all",
        ShowPath.ARG_LONG_NAME: "full",
        StrictOrPermissive.PERMISSIVE_ARG_LONG_NAME: 1,
        Output.SHOW_IR_ARG_LONG_NAME: True,
        Output.SHOW_RESULTS_ARG_LONG_NAME: True,
        Output.SHOW_STATS_ARG_LONG_NAME: True,
        Cache.ARG_LONG_NAME: "rattr/cli/parser.py",
    }, (
        f"Error parsing pyproject.toml. Arg: '{ExcludePatterns.ARG_LONG_NAME}' "
        f"is of wrong type: 'bool'. Expected type: "
        f"'{ExcludePatterns.TOML_ARG_NAME_ARG_TYPE_MAP[ExcludePatterns.ARG_LONG_NAME].__name__}'."  # noqa: E501
    )


@pytest.fixture
def toml_dicts_with_bad_field_types_and_errors(
    toml_dict_with_bad_field_types_and_error1,
    toml_dict_with_bad_field_types_and_error2,
    toml_dict_with_bad_field_types_and_error3,
):
    return [
        toml_dict_with_bad_field_types_and_error1,
        toml_dict_with_bad_field_types_and_error2,
        toml_dict_with_bad_field_types_and_error3,
    ]


@pytest.fixture
def toml_dict_with_mutex_arg_violation_and_error1():
    return {
        FollowImports.ARG_LONG_NAME: 3,
        ExcludeImports.ARG_LONG_NAME: [
            "a\\.b\\.c",
            "a\\.b.*",
            "a\\.b\\.c\\.e",
            "a\\.b\\.c.*",
        ],
        ExcludePatterns.ARG_LONG_NAME: ["a_.*", "b_.*", "_.*"],
        ShowWarnings.ARG_LONG_NAME: "all",
        ShowPath.ARG_LONG_NAME: "full",
        StrictOrPermissive.PERMISSIVE_ARG_LONG_NAME: 1,
        Output.SHOW_IR_ARG_LONG_NAME: True,  # mutex violation
        Output.SILENT_ARG_LONG_NAME: True,  # mutex violation
        Cache.ARG_LONG_NAME: "cache.json",
    }, (
        f"-irs "
        f"('--{Output.SHOW_IR_ARG_LONG_NAME}', "
        f"'--{Output.SHOW_RESULTS_ARG_LONG_NAME}', "
        f"'--{Output.SHOW_STATS_ARG_LONG_NAME}') "
        f"and -S ('--{Output.SILENT_ARG_LONG_NAME}') are mutually exclusive"
    )


@pytest.fixture
def toml_dict_with_mutex_arg_violation_and_error2():
    return {
        FollowImports.ARG_LONG_NAME: 3,
        ExcludeImports.ARG_LONG_NAME: [
            "a\\.b\\.c",
            "a\\.b.*",
            "a\\.b\\.c\\.e",
            "a\\.b\\.c.*",
        ],
        ExcludePatterns.ARG_LONG_NAME: ["a_.*", "b_.*", "_.*"],
        ShowWarnings.ARG_LONG_NAME: "all",
        ShowPath.ARG_LONG_NAME: "full",
        StrictOrPermissive.STRICT_ARG_LONG_NAME: True,  # mutex violation
        StrictOrPermissive.PERMISSIVE_ARG_LONG_NAME: 1,  # mutex violation
        Output.SHOW_IR_ARG_LONG_NAME: True,
        Output.SHOW_RESULTS_ARG_LONG_NAME: True,
        Output.SHOW_STATS_ARG_LONG_NAME: True,
        Cache.ARG_LONG_NAME: "cache.json",
    }, (
        f"argument --{StrictOrPermissive.PERMISSIVE_ARG_LONG_NAME}: not allowed "
        f"with argument --{StrictOrPermissive.STRICT_ARG_LONG_NAME}"
    )


@pytest.fixture
def toml_dicts_with_mutex_arg_violation_and_errors(
    toml_dict_with_mutex_arg_violation_and_error1,
    toml_dict_with_mutex_arg_violation_and_error2,
):
    return [
        toml_dict_with_mutex_arg_violation_and_error1,
        toml_dict_with_mutex_arg_violation_and_error2,
    ]


@pytest.fixture
def correct_toml_dict1():
    return {
        FollowImports.ARG_LONG_NAME: 3,
        ExcludeImports.ARG_LONG_NAME: [
            "a\\.a\\.a",
            "b\\.b.*",
            "c\\.c\\.c\\.c",
            "d\\d\\.d.*",
        ],
        ExcludePatterns.ARG_LONG_NAME: ["a_.*", "b_.*", "c_.*"],
        ShowWarnings.ARG_LONG_NAME: "all",
        StrictOrPermissive.STRICT_ARG_LONG_NAME: True,
        Cache.ARG_LONG_NAME: "cache.json",
    }


@pytest.fixture
def correct_toml_dict2():
    return {
        FollowImports.ARG_LONG_NAME: 3,
        ExcludeImports.ARG_LONG_NAME: [
            "a\\.a\\.a",
            "b\\.b.*",
            "c\\.c\\.c\\.c",
            "d\\d\\.d.*",
        ],
        ExcludePatterns.ARG_LONG_NAME: ["a_.*", "b_.*", "c_.*"],
        ShowWarnings.ARG_LONG_NAME: "all",
        ShowPath.ARG_LONG_NAME: "full",
        StrictOrPermissive.PERMISSIVE_ARG_LONG_NAME: 1,
        Cache.ARG_LONG_NAME: "cache.json",
    }


@pytest.fixture
def correct_toml_dict3():
    return {
        FollowImports.ARG_LONG_NAME: 1,
        ExcludeImports.ARG_LONG_NAME: [
            "a\\.b\\.c",
            "a\\.b.*",
            "a\\.b\\.c\\.e",
            "a\\.b\\.c.*",
        ],
        ShowWarnings.ARG_LONG_NAME: "all",
        ShowPath.ARG_LONG_NAME: "full",
        StrictOrPermissive.STRICT_ARG_LONG_NAME: True,
        Output.SHOW_IR_ARG_LONG_NAME: True,
        Output.SHOW_RESULTS_ARG_LONG_NAME: True,
        Output.SHOW_STATS_ARG_LONG_NAME: True,
        Cache.ARG_LONG_NAME: "cache.json",
    }


@pytest.fixture
def sys_args1():
    return [
        "rattr",
        "rattr/cli/parser.py",
    ]


@pytest.fixture
def sys_args2():
    return [
        "rattr",
        "--follow-imports",
        "2",
        "--strict",
        "-w",
        "file",
        "rattr/cli/argparse.py",
        "-p",
        "short",
    ]


@pytest.fixture
def sys_args3():
    return [
        "rattr",
        "--follow-imports",
        "2",
        "--strict",
        "-w",
        "file",
        "rattr/cli/argparse.py",
        "-p",
        "short",
        "-F",
        "*exclude-import*",
        "-x",
        "*exclude-pattern*",
        "-S",
    ]


@pytest.fixture
def sys_args4():
    return ["rattr", "rattr/cli/parser.py", "-S"]


@pytest.fixture
def sys_args5():
    return ["rattr", "rattr/cli/parser.py", "--permissive", "3"]


@pytest.fixture
def walrus():
    def parse(expr: str):
        return ast.parse(f"a = ({expr})").body[0].value

    def _inner(*exprs):
        if len(exprs) == 0:
            raise ValueError

        if len(exprs) == 1:
            return parse(exprs[0])
        else:
            return [parse(e) for e in exprs]

    return _inner


@pytest.fixture
def stringify_nodes():
    def _inner(nodes: ast.AST):
        return [ast.dump(n) for n in nodes]

    return _inner



#!/usr/bin/env python3
"""Rattr entry point."""

import json
from math import log10
from typing import Iterable, Set

from rattr import config, error
from rattr.analyser.context import Import, Symbol
from rattr.analyser.file import RattrStats, parse_and_analyse_file
from rattr.analyser.results import ResultsEncoder, generate_results_from_ir
from rattr.analyser.types import FileIR, FileResults, ImportsIR
from rattr.analyser.util import (
    cache_is_valid,
    create_cache,
    is_blacklisted_module,
    re_filter_ir,
    re_filter_results,
)
from rattr.cli import Namespace, parse_arguments
from rattr.error import get_badness


def main(arguments: Namespace) -> None:
    """Rattr entry point."""
    load_config(arguments)

    file_ir, imports_ir, stats = parse_and_analyse_file()

    results = generate_results_from_ir(file_ir, imports_ir)

    if not error.is_within_badness_threshold():
        error.fatal(f"exceeded allowed badness ({get_badness()} > {config.threshold})")

    if config.show_ir:
        show_ir(config.file, file_ir, imports_ir)

    if config.show_results:
        show_results(results)

    if config.show_stats:
        show_stats(stats)

    if config.cache:
        write_cache(results, file_ir, imports_ir)


def show_ir(file: str, file_ir: FileIR, imports_ir: ImportsIR) -> None:
    """Prettily print the given file and imports IR."""
    if config.filter_string:
        imports_ir = {
            i: re_filter_ir(ir, config.filter_string) for i, ir in imports_ir.items()
        }
        file_ir = re_filter_ir(file_ir, config.filter_string)

    jsonable_ir = dict()

    for i, ir in imports_ir.items():
        jsonable_ir[i] = {repr(c): r for c, r in ir.items()}

    jsonable_ir[file] = {repr(c): r for c, r in file_ir.items()}

    print(json.dumps(jsonable_ir, indent=4, cls=ResultsEncoder))


def show_results(results: FileResults) -> None:
    """Prettily print the given file results."""
    if config.filter_string:
        results = re_filter_results(results, config.filter_string)

    print(json.dumps(results, indent=4, cls=ResultsEncoder))


def show_stats(stats: RattrStats) -> None:
    """Prettily print the collected stats in tables."""
    row = "{:26} | {:18}"

    # Collate time stats
    table_header = row.format("", "Time (Seconds)")
    table_width = len(table_header)
    times = {
        "Parse <file>": stats.parse_time,
        "Build root context": stats.root_context_time,
        "Run assertor": stats.assert_time,
        "Parse / analyse imports": stats.analyse_imports_time,
        "Analyse <file>": stats.analyse_file_time,
    }

    # Print time stats
    print(end="\n\n")
    print(table_header)
    print("=" * table_width)
    for col_one, col_two in times.items():
        print(row.format(col_one, format(col_two, ".9f") + " s"))

    # Collate imports stats
    table_header = row.format("", "# of Imports")
    table_width = len(table_header)
    imports = {
        "Total number of imports": stats.number_of_imports,
        "... of which unique": stats.number_of_unique_imports,
    }

    # Print imports stats
    print(end="\n\n")
    print(table_header)
    print("=" * table_width)
    digits = 1 + int(log10(max(*imports.values(), 1)))
    for col_one, col_two in imports.items():
        print(row.format(col_one, format(col_two).zfill(digits)))

    # Collate line stats
    table_header = row.format("", "# of Lines")
    table_width = len(table_header)
    lines = {
        "<file>": stats.file_lines,
        "Imports": stats.import_lines,
    }

    # Print line stats
    print(end="\n\n")
    print(table_header)
    print("=" * table_width)
    digits = 1 + int(log10(max(*lines.values())))
    for col_one, col_two in lines.items():
        print(row.format(col_one, format(col_two).zfill(digits)))

    # Collate badness stats
    table_header = row.format("", "Magnitude")
    table_width = len(table_header)
    badness_stats = {
        "Total badness": sum(
            (config.file_badness, config.import_badness, config.simplify_badness)
        ),
        "... from <file>": config.file_badness,
        "... from imports": config.import_badness,
        "... from simplification": config.simplify_badness,
    }

    # Print badness stats
    print(end="\n\n")
    print(table_header)
    print("=" * table_width)
    if max(*badness_stats.values()) > 0:
        digits = 1 + int(log10(max(*badness_stats.values())))
    else:
        digits = 1
    for col_one, col_two in badness_stats.items():
        print(row.format(col_one, format(col_two).zfill(digits)))

    # Badness summary
    badness = config.file_badness + config.simplify_badness
    badness_summary_stats = {
        "True badness": badness,
        "Threshold": config.threshold if config.threshold > 0 else "∞",
        "Average badness/line": format(badness / stats.file_lines, ".3f")
        + " b/l (true badness / <file> lines)",
    }
    summary = f"{{:{1 + max(map(len, badness_summary_stats.keys()))}}}: {{}}"
    print(end="\n\n")
    for desc, stat in badness_summary_stats.items():
        print(summary.format(desc, stat))

    # Summary stats
    summary_stats = {
        "Total time": format(sum(times.values()), ".6f") + " s",
        "Total lines": sum(lines.values()),
        "Average lines/second": format(sum(lines.values()) / sum(times.values()), ".0f")
        + " l/s",
    }
    summary = f"{{:{1 + max(map(len, summary_stats.keys()))}}}: {{}}"
    print(end="\n\n")
    for desc, stat in summary_stats.items():
        print(summary.format(desc, stat))

    print(end="\n\n")


def write_cache(results: FileResults, file_ir: FileIR, imports_ir: ImportsIR) -> None:
    """Save the file results to the file cache."""
    if cache_is_valid(config.file, config.cache):
        return error.rattr(f"cache for '{config.file}' is already up to date")

    imports: Set[str] = set()

    for ctx in [file_ir.context, *[i.context for i in imports_ir.values()]]:
        symbols: Iterable[Symbol] = ctx.symbol_table.symbols()
        for sym in symbols:
            if not isinstance(sym, Import):
                continue

            if is_blacklisted_module(sym.module_name):
                continue

            if sym.module_spec is None or sym.module_spec.origin is None:
                continue

            if sym.module_spec.origin == "built-in":
                continue

            imports.add(sym.module_spec.origin)

    create_cache(results, imports, ResultsEncoder)


def load_config(arguments: Namespace) -> None:
    """Populate the config with the given arguments."""
    config.follow_imports = arguments.follow_imports
    config.follow_pip_imports = arguments.follow_imports >= 2
    config.follow_stdlib_imports = arguments.follow_imports >= 3

    config.excluded_imports = set(arguments.exclude_import)
    config.excluded_names = set(arguments.exclude)

    config.show_warnings = arguments.show_warnings != "none"
    config.show_imports_warnings = arguments.show_warnings in ("all", "ALL")
    config.show_low_priority_warnings = arguments.show_warnings == "ALL"

    config.show_path = arguments.show_path != "none"
    config.use_short_path = arguments.show_path == "short"

    config.strict = arguments.strict
    config.permissive = arguments.permissive
    config.threshold = arguments.threshold

    config.show_ir = arguments.show_ir
    config.show_results = arguments.show_results
    config.show_stats = arguments.show_stats
    config.silent = arguments.silent

    config.filter_string = arguments.filter_string
    config.file = arguments.file

    config.cache = arguments.cache


def entry_point():
    """Entry point for command line app."""
    main(parse_arguments())


if __name__ == "__main__":
    entry_point()



"""Base classes for Rattr components."""

import ast
from abc import ABCMeta, abstractmethod, abstractproperty
from ast import NodeTransformer, NodeVisitor  # noqa: F401
from itertools import product
from typing import Dict, List, Optional

from rattr import error
from rattr.analyser.context import Context
from rattr.analyser.context.symbol import Import
from rattr.analyser.types import FuncOrAsyncFunc, FunctionIR


class Assertor(NodeVisitor):
    """Assertor base class.

    An assertor can either be strict or non-strict, set by the constructor
    argument `is_strict`, having the following affect:
        is_strict:
            on condition failure, log fatal error (halts execution)
        not is_strict:
            on condition failure, log warning

    """

    def __init__(self, is_strict: bool = True) -> None:
        self.is_strict: bool = is_strict
        self.context: Context = None

    def assert_holds(self, node: ast.AST, context: Context) -> None:
        """Entry point for an Assertor, visit the tree to assert properties."""
        self.context = context

        super().visit(node)

    def failed(self, message: str, culprit: Optional[ast.AST] = None) -> None:
        """Handle assertion failure."""
        if self.is_strict:
            handler = error.fatal
        else:
            handler = error.warning

        handler(message, culprit)


class CustomFunctionAnalyser(NodeVisitor, metaclass=ABCMeta):
    """Base class for a custom function visitor."""

    @abstractproperty
    def name(self) -> str:
        """Return the name of the function handled by this analyser."""
        return ""

    @abstractproperty
    def qualified_name(self) -> str:
        """Return the qualified name of the function."""
        return ""

    @abstractmethod
    def on_def(self, name: str, node: FuncOrAsyncFunc, ctx: Context) -> FunctionIR:
        """Return the IR of the definition of the handled function."""
        return {
            "sets": set(),
            "gets": set(),
            "dels": set(),
            "calls": set(),
        }

    @abstractmethod
    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIR:
        """Return the IR produced by a call to the handled function.

        The returned IR will be union'd with the IR of the caller function.

        Argument `fna` is the FunctionAnalyser instance that the call is from.

        """
        return {
            "sets": set(),
            "gets": set(),
            "dels": set(),
            "calls": set(),
        }


class CustomFunctionHandler:
    """Dispatch to CustomFunctionAnalyser.

    If a builtin and user-defined function have the same name, then the builtin
    function analyser will take precedence.

    Initialiser:
        Register the given builtin and user-defined custom function analysers.

    """

    def __init__(
        self,
        builtins: Optional[List[CustomFunctionAnalyser]] = None,
        user_defined: Optional[List[CustomFunctionAnalyser]] = None,
    ) -> None:
        self._builtins: Dict[str, CustomFunctionAnalyser] = dict()
        self._user_def: Dict[str, CustomFunctionAnalyser] = dict()

        for analyser in builtins or []:
            self._builtins[analyser.name] = analyser

        for analyser in user_defined or []:
            self._user_def[analyser.name] = analyser

    def __get_by_name(self, name: str) -> Optional[CustomFunctionAnalyser]:
        analyser = None

        if name in self._user_def:
            analyser = self._user_def[name]

        if name in self._builtins:
            analyser = self._builtins[name]

        return analyser

    def __get_by_symbol(
        self, name: str, ctx: Context
    ) -> Optional[CustomFunctionAnalyser]:
        symbols: List[Import] = list()

        _imports: filter[Import] = filter(
            lambda s: isinstance(s, Import), ctx.symbol_table.symbols()
        )

        for symbol in _imports:
            # From imported
            if name in (symbol.name, symbol.qualified_name):
                symbols.append(symbol)

            # Module imported
            if name.startswith(f"{symbol.name}."):
                symbols.append(symbol)

        for symbol, analyser in product(symbols, self._user_def.values()):
            # From imported
            if analyser.qualified_name == symbol.qualified_name:
                return analyser

            # Module imported
            if analyser.qualified_name == name.replace(
                symbol.name, symbol.qualified_name
            ):
                return analyser

        return None

    def get(self, name: str, ctx: Context) -> Optional[CustomFunctionAnalyser]:
        """Return the analyser for the function `name`, `None` otherwise."""
        analyser = self.__get_by_name(name)

        if analyser is None:
            analyser = self.__get_by_symbol(name, ctx)

        return analyser

    def has_analyser(self, name: str, ctx: Context) -> bool:
        """Return `True` if there is a analyser for the function `name`."""
        return self.get(name, ctx) is not None

    def handle_def(self, name: str, node: FuncOrAsyncFunc, ctx: Context) -> FunctionIR:
        """Dispatch to the to the appropriate analyser."""
        analyser = self.get(name, ctx)

        if analyser is None:
            raise ValueError

        return analyser.on_def(name, node, ctx)

    def handle_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIR:
        """Dispatch to the to the appropriate analyser."""
        analyser = self.get(name, ctx)

        if analyser is None:
            raise ValueError

        return analyser.on_call(name, node, ctx)



import pytest

from rattr.analyser.context import Call, Func, Name, RootContext
from rattr.analyser.file import FileAnalyser
from rattr.plugins import plugins
from rattr.plugins.analysers.builtins import SortedAnalyser


class TestCustomFunctionAnalysers:
    @pytest.fixture(autouse=True)
    def apply_plugins(self):
        plugins.register(SortedAnalyser())

    def test_sorted_no_key(self, parse):
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "gets": {
                    Name("arg"),
                },
                "sets": set(),
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

    def test_sorted_with_constant_key(self, parse):
        # Literal
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg, key=1)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "gets": {
                    Name("arg"),
                },
                "sets": set(),
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

        # Attribute
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg, key=arg.attr)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "gets": {
                    Name("arg"),
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

    def test_sorted_key_is_callable(self, parse):
        # Via lambda
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg, key=lambda a: key_func(a))

            def key_func(a):
                return a.b
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("a_func", ["arg"], None, None)
        key_func = Func("key_func", ["a"], None, None)
        expected = {
            a_func: {
                "gets": {
                    Name("arg"),
                },
                "sets": set(),
                "dels": set(),
                "calls": {Call("key_func()", ["a"], {}, target=key_func)},
            },
            key_func: {
                "gets": {
                    Name("a.b", "a"),
                },
                "sets": set(),
                "dels": set(),
                "calls": set(),
            },
        }

        assert results == expected

    def test_sorted_key_is_attribute_or_element(self, parse):
        # Attribute
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg, key=lambda a: a.attr)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "gets": {
                    Name("arg"),
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

        # Element
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg, key=lambda a: a[0])
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "gets": {
                    Name("arg"),
                    Name("arg[]", "arg"),
                },
                "sets": set(),
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

        # Complex
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg, key=lambda a: a[0].attr)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "gets": {
                    Name("arg"),
                    Name("arg[].attr", "arg"),
                },
                "sets": set(),
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected



from unittest import mock

import pytest

from rattr.analyser.context import Builtin, Call, Func, Import, Name
from rattr.analyser.context.symbol import Class
from rattr.analyser.ir_dag import (
    IrDagNode,
    construct_swap,
    get_callee_target,
    partially_unbind,
    partially_unbind_name,
)


class TestIrDag_Utils:
    def test_get_callee_target(self):
        fn_a = Func("fn_a", [], None, None)
        fn_b = Func("fn_b", [], None, None)
        cls_a = Class("ClassA", ["self", "arg"], None, None)
        fn_a_ir = {
            "sets": {Name("a"), Name("b.attr", "b")},
            "gets": {Name("c")},
            "dels": set(),
            "calls": set(),
        }
        fn_b_ir = {
            "sets": set(),
            "gets": set(),
            "dels": {Name("a")},
            "calls": set(),
        }
        cls_a_ir = {
            "sets": {Name("self.field", "self")},
            "gets": {Name("arg.attr", "arg")},
            "dels": set(),
            "calls": set(),
        }
        file_ir = {
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
            cls_a: cls_a_ir,
        }

        # No callee target
        broken_call = Call("some_func()", [], {}, target=None)
        assert get_callee_target(broken_call, file_ir, {}) == (None, None)

        # Callee is a builtin
        builtin = Call("max()", [], {}, Builtin("max", has_affect=False))
        assert get_callee_target(builtin, file_ir, {}) == (None, None)

        # Normal -- `fn_a`
        fn_a_call = Call("fn_a()", [], {}, target=fn_a)
        assert get_callee_target(fn_a_call, file_ir, {}) == (fn_a, fn_a_ir)

        # Normal -- `fn_b`
        fn_b_call = Call("fn_b()", [], {}, target=fn_b)
        assert get_callee_target(fn_b_call, file_ir, {}) == (fn_b, fn_b_ir)

        # Normal -- class initialiser
        cls_a_call = Call("ClassA()", ["a"], {}, target=cls_a)
        assert get_callee_target(cls_a_call, file_ir, {}) == (cls_a, cls_a_ir)

        # Non-existent
        fake_call = Call("not_a_real_function()", [], {})
        assert get_callee_target(fake_call, file_ir, {}) == (None, None)

    @pytest.mark.pypy()
    def test_get_callee_target_callee_in_stdlib(self):
        fn_a = Func("fn_a", [], None, None)
        fn_b = Func("fn_b", [], None, None)
        cls_a = Class("ClassA", ["self", "arg"], None, None)
        fn_a_ir = {
            "sets": {Name("a"), Name("b.attr", "b")},
            "gets": {Name("c")},
            "dels": set(),
            "calls": set(),
        }
        fn_b_ir = {
            "sets": set(),
            "gets": set(),
            "dels": {Name("a")},
            "calls": set(),
        }
        cls_a_ir = {
            "sets": {Name("self.field", "self")},
            "gets": {Name("arg.attr", "arg")},
            "dels": set(),
            "calls": set(),
        }
        file_ir = {
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
            cls_a: cls_a_ir,
        }

        # Callee is in stdlib
        stdlib = Call("sin()", [], {}, target=Import("sin", "math.sin"))
        assert stdlib.target.module_name == "math"
        assert get_callee_target(stdlib, file_ir, {}) == (None, None)

    def test_get_callee_target_imported_function(self, file_ir_from_dict, capfd):
        # TODO Imported class/method
        fn = Func("fn", [], None, None)
        fn_ir = {"sets": set(), "gets": set(), "dels": set(), "calls": set()}
        imports_ir = {"module": file_ir_from_dict({fn: fn_ir})}

        # Callee is imported
        _i = Import("fn", "module.fn")
        _i.module_name = "module"
        _i.module_spec = mock.Mock()
        call = Call("fn()", [], {}, target=_i)
        assert get_callee_target(call, {}, imports_ir) == (fn, fn_ir)

        # Callee is imported but not defined in source
        _i = Import("nope", "module.nope")
        _i.module_name = "module"
        _i.module_spec = mock.Mock()
        call = Call("nope()", [], {}, target=_i)

        assert get_callee_target(call, {}, imports_ir) == (None, None)

        output, _ = capfd.readouterr()
        assert "unable to resolve call to 'nope' in import" in output

        # Callee module and target not found
        _i = Import("nah", "noway.nah")
        _i.module_name = "noway"
        _i.module_spec = mock.Mock()
        call = Call("nah()", [], {}, target=_i)

        with pytest.raises(ImportError):
            get_callee_target(call, {}, imports_ir)

    def test_partially_unbind_name(self):
        # On basic name
        foo, star_foo = Name("foo"), Name("*foo", "foo")

        assert partially_unbind_name(foo, "foo") == foo
        assert partially_unbind_name(star_foo, "foo") == star_foo

        assert partially_unbind_name(foo, "bar") == Name("bar")
        assert partially_unbind_name(star_foo, "bar") == Name("*bar", "bar")

        # On complex name
        comp = Name("comp.attr.m().res_attr[].item", "comp")
        star_comp = Name("*comp.attr.m().res_attr[].item", "comp")

        assert partially_unbind_name(comp, "comp") == comp
        assert partially_unbind_name(star_comp, "comp") == star_comp

        expected_comp = Name("bar.attr.m().res_attr[].item", "bar")
        expected_star_comp = Name("*bar.attr.m().res_attr[].item", "bar")

        assert partially_unbind_name(comp, "bar") == expected_comp
        assert partially_unbind_name(star_comp, "bar") == expected_star_comp

    def test_partially_unbind(self):
        empty_func_ir = {
            "sets": set(),
            "gets": set(),
            "dels": set(),
            "calls": set(),
        }
        func_ir = {
            "sets": {
                Name("a"),
                Name("a.attr", "a"),
            },
            "gets": {
                Name("arg"),
                Name("arg.mth().res_attr[].value", "arg"),
            },
            "dels": {
                Name("bob"),
                Name("*dob", "dob"),
            },
            "calls": {Call("callee()", [], {})},
        }

        # No names, no swaps
        assert partially_unbind(empty_func_ir, {}) == empty_func_ir

        # No names, w/ swaps
        assert partially_unbind(empty_func_ir, {"a": "b"}) == empty_func_ir

        # No swaps
        assert partially_unbind(func_ir, {}) == func_ir

        # Only relevant swaps #1
        expected = {
            "sets": {
                Name("b"),
                Name("b.attr", "b"),
            },
            "gets": {
                Name("arg"),
                Name("arg.mth().res_attr[].value", "arg"),
            },
            "dels": {
                Name("bob"),
                Name("*dob", "dob"),
            },
            "calls": {Call("callee()", [], {})},
        }
        swaps = {"a": "b"}

        assert partially_unbind(func_ir, swaps) == expected

        # Only relevant swaps #2
        expected = {
            "sets": {
                Name("a"),
                Name("a.attr", "a"),
            },
            "gets": {
                Name("barg"),
                Name("barg.mth().res_attr[].value", "barg"),
            },
            "dels": {
                Name("bobby"),
                Name("*dob", "dob"),
            },
            "calls": {Call("callee()", [], {})},
        }
        swaps = {
            "arg": "barg",
            "bob": "bobby",
        }

        assert partially_unbind(func_ir, swaps) == expected

        # Mixed swaps
        expected = {
            "sets": {
                Name("a"),
                Name("a.attr", "a"),
            },
            "gets": {
                Name("arg"),
                Name("arg.mth().res_attr[].value", "arg"),
            },
            "dels": {
                Name("bob"),
                Name("*dib", "dib"),
            },
            "calls": {Call("callee()", [], {})},
        }
        swaps = {"dob": "dib", "xyz": "zyx"}

        assert partially_unbind(func_ir, swaps) == expected

        # Regression
        func_ir = {
            "sets": {
                Name(name="arg_b.set_in_fn_b", basename="arg_b"),
            },
            "gets": set(),
            "dels": set(),
            "calls": set(),
        }
        expected = {
            "sets": {
                Name(name="arg.set_in_fn_b", basename="arg"),
            },
            "gets": set(),
            "dels": set(),
            "calls": set(),
        }
        swaps = {"arg_b": "arg"}

        assert partially_unbind(func_ir, swaps) == expected

    def test_construct_swap(self, capfd):
        # No args in both
        fn_def = Func("fn", [], None, None)
        fn_call = Call("fn", [], {})

        expected = {}

        assert construct_swap(fn_def, fn_call) == expected

        # Positional
        fn_def = Func("fn", ["a", "b"], None, None)
        fn_call = Call("fn", ["c", "d"], {})

        expected = {"a": "c", "b": "d"}

        assert construct_swap(fn_def, fn_call) == expected

        # Keyword
        fn_def = Func("fn", ["a", "b"], None, None)
        fn_call = Call("fn", [], {"a": "not_a", "b": "not_b"})

        expected = {"a": "not_a", "b": "not_b"}

        assert construct_swap(fn_def, fn_call) == expected

        # Mixed
        fn_def = Func("fn", ["a", "b", "c", "d"], None, None)
        fn_call = Call("fn", ["not_a", "not_b"], {"d": "not_d", "c": "not_c"})

        expected = {"a": "not_a", "b": "not_b", "c": "not_c", "d": "not_d"}

        assert construct_swap(fn_def, fn_call) == expected

        # Complex mixed w/ default on 'e'
        fn_def = Func("fn", ["a", "b", "c", "d", "e"], None, None)
        fn_call = Call("fn", ["not_a", "not_b"], {"d": "not_d", "c": "not_c"})

        expected = {"a": "not_a", "b": "not_b", "c": "not_c", "d": "not_d"}

        assert construct_swap(fn_def, fn_call) == expected

        # No matches
        fn_def = Func("fn", ["a", "b"], None, None)
        fn_call = Call("fn", [], {"c": "no_match", "d": "no_match"})

        expected = {}

        assert construct_swap(fn_def, fn_call) == expected

        output, _ = capfd.readouterr()

        assert "unexpected named arguments" in output

        # Provide as positional and named
        fn_def = Func("fn", ["a"], None, None)
        fn_call = Call("fn", ["not_a"], {"a": "not_a"})

        with mock.patch("sys.exit") as _exit:
            construct_swap(fn_def, fn_call)

        assert _exit.call_count == 1


class TestIrDagNode:

    # NOTE
    #   TestIrDagNode.simplify tested in the test for `results.py`'s
    #   `generate_results_from_ir` function as the tests would be almost
    #   identical either way

    def test_populate(self):
        fn_a = Func("fn_a", ["a"], None, None)
        fn_b = Func("fn_b", ["b"], None, None)
        fn_c = Func("fn_c", ["c"], None, None)
        fn_d = Func("fn_d", ["d"], None, None)
        fn_a_ir = {
            "sets": {Name("a")},
            "gets": set(),
            "dels": set(),
            "calls": {Call("fn_b()", ["a"], {}, target=fn_b)},
        }
        fn_b_ir = {
            "sets": {Name("b")},
            "gets": set(),
            "dels": set(),
            "calls": {
                Call("fn_c()", ["b"], {}, target=fn_c),
                Call("fn_d()", ["b"], {}, target=fn_d),
            },
        }
        fn_c_ir = {
            "sets": {Name("c")},
            "gets": set(),
            "dels": set(),
            "calls": set(),
        }
        fn_d_ir = {
            "sets": {Name("d")},
            "gets": set(),
            "dels": set(),
            "calls": set(),
        }
        file_ir = {
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
            fn_c: fn_c_ir,
            fn_d: fn_d_ir,
        }

        A = IrDagNode(None, fn_a, fn_a_ir, file_ir, dict())
        A.populate()

        B_in_A = A.children[0]

        # C and D can be either way around as it is a set
        if B_in_A.children[0].func.name == fn_c.name:
            C_in_A = B_in_A.children[0]
            D_in_A = B_in_A.children[1]
        else:
            C_in_A = B_in_A.children[1]
            D_in_A = B_in_A.children[0]

        # A
        assert A.call is None
        assert A.func == fn_a
        assert A.func_ir == fn_a_ir
        assert A.file_ir == file_ir
        assert len(A.children) == 1

        # B
        assert B_in_A.call == Call("fn_b()", ["a"], {}, target=fn_b)
        assert B_in_A.func == fn_b
        assert B_in_A.func_ir == fn_b_ir
        assert B_in_A.file_ir == file_ir
        assert len(B_in_A.children) == 2

        # C
        assert C_in_A.call == Call("fn_c()", ["b"], {}, target=fn_c)
        assert C_in_A.func == fn_c
        assert C_in_A.func_ir == fn_c_ir
        assert C_in_A.file_ir == file_ir
        assert len(C_in_A.children) == 0

        # D
        assert D_in_A.call == Call("fn_d()", ["b"], {}, target=fn_d)
        assert D_in_A.func == fn_d
        assert D_in_A.func_ir == fn_d_ir
        assert D_in_A.file_ir == file_ir
        assert len(D_in_A.children) == 0

    def test_on_undefined(self, capfd):
        fn_a = Func("fn_a", ["a"], None, None)
        fn_a_ir = {
            "sets": {Name("a")},
            "gets": set(),
            "dels": set(),
            "calls": {
                Call("fn_b()", ["a"], {}),
                Call("some_undefined_func()", [], {}),
            },
        }
        fn_b = Func("fn_b", ["b"], None, None)
        fn_b_ir = {
            "sets": {Name("b")},
            "gets": set(),
            "dels": set(),
            "calls": {
                Call("some_other_undefined_func()", [], {}),
            },
        }
        file_ir = {
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
        }

        A = IrDagNode(None, fn_a, fn_a_ir, file_ir, {})

        # NOTE Error is already logged by analyser, thus nothing should happen
        A.populate()
        A.simplify()

        output, _ = capfd.readouterr()

        assert output == ""

    def test_populate_on_stdlib(self, capfd):
        fn_a = Func("fn_a", ["a"], None, None)
        fn_a_ir = {
            "sets": {Name("a")},
            "gets": set(),
            "dels": set(),
            "calls": {
                Call("os.path.join()", [], {}),
                Call("math.max()", [], {}),
            },
        }
        file_ir = {
            fn_a: fn_a_ir,
        }

        A = IrDagNode(None, fn_a, fn_a_ir, file_ir, dict())
        A.populate()

        output, _ = capfd.readouterr()

        assert output == ""

    def test_populate_ignore_builtins(self, capfd):
        fn_a = Func("fn_a", ["a"], None, None)
        fn_a_ir = {
            "sets": {Name("a")},
            "gets": set(),
            "dels": set(),
            "calls": {
                Call("max()", [], {}),
                Call("enumerate()", [], {}),
            },
        }
        file_ir = {
            fn_a: fn_a_ir,
        }

        A = IrDagNode(None, fn_a, fn_a_ir, file_ir, dict())
        A.populate()

        output, _ = capfd.readouterr()

        assert output == ""

    def test_populate_ignore_seen(self):
        fn_a = Func("fn_a", ["a"], None, None)
        fn_b = Func("fn_b", ["b"], None, None)
        fn_c = Func("fn_c", ["c"], None, None)
        fn_a_ir = {
            "sets": {Name("a")},
            "gets": set(),
            "dels": set(),
            "calls": {
                Call("fn_a()", ["a"], {}, target=fn_a),
                Call("fn_b()", ["a"], {}, target=fn_b),
            },
        }
        fn_b_ir = {
            "sets": {Name("b")},
            "gets": set(),
            "dels": set(),
            "calls": {
                Call("fn_a()", ["b"], {}, target=fn_a),
                Call("fn_c()", ["b"], {}, target=fn_c),
            },
        }
        fn_c_ir = {
            "sets": {Name("b")},
            "gets": set(),
            "dels": set(),
            "calls": {
                Call("fn_b()", ["c"], {}, target=fn_b),
            },
        }
        file_ir = {
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
            fn_c: fn_c_ir,
        }

        A = IrDagNode(None, fn_a, fn_a_ir, file_ir, dict())
        A.populate()

        # Can be either way around as it is a set
        if A.children[0].func.name == fn_b.name:
            B_in_A = A.children[0]
        else:
            B_in_A = A.children[1]

        # Can be either way around as it is a set
        if B_in_A.children[0].func.name == fn_c.name:
            C_in_A = B_in_A.children[0]
        else:
            C_in_A = B_in_A.children[1]

        # A
        assert A.call is None
        assert A.func == fn_a
        assert A.func_ir == fn_a_ir
        assert A.file_ir == file_ir
        assert len(A.children) == 2

        # B
        assert B_in_A.call == Call("fn_b()", ["a"], {}, target=fn_b)
        assert B_in_A.func == fn_b
        assert B_in_A.func_ir == fn_b_ir
        assert B_in_A.file_ir == file_ir
        assert len(B_in_A.children) == 2

        # C
        assert C_in_A.call == Call("fn_c()", ["b"], {}, target=fn_c)
        assert C_in_A.func == fn_c
        assert C_in_A.func_ir == fn_c_ir
        assert C_in_A.file_ir == file_ir
        assert len(C_in_A.children) == 1



import pytest

from rattr.analyser.context import (
    Builtin,
    Call,
    Class,
    Func,
    Name,
    RootContext,
)
from rattr.analyser.file import FileAnalyser


class TestFunctionAnalyser:
    def test_basic_function(self, parse):
        _ast = parse(
            """
            def a_func(arg):
                arg.sets_me = "value"
                return arg.gets_me
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {Name("arg.gets_me", "arg")},
                "sets": {Name("arg.sets_me", "arg")},
            }
        }

        assert results == expected

    def test_multiple_functions(self, parse):
        _ast = parse(
            """
            def a_func(arg):
                arg.sets_me = "value"
                return arg.gets_me

            def another_func(arg):
                arg.attr = "this function only sets"
                arg.attr_two = "see!"
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {Name("arg.gets_me", "arg")},
                "sets": {Name("arg.sets_me", "arg")},
            },
            Func("another_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": set(),
                "sets": {
                    Name("arg.attr", "arg"),
                    Name("arg.attr_two", "arg"),
                },
            },
        }

        assert results == expected

    def test_conditional(self, parse):
        _ast = parse(
            """
            def a_func(arg):
                if debug:
                    return False
                return arg.attr == "target value"
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("debug"),
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

    def test_nested_conditional(self, parse):
        _ast = parse(
            """
            def a_func(arg):
                if c1:
                    if c2:
                        return arg.foo
                return arg.bar
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("c1"),
                    Name("c2"),
                    Name("arg.foo", "arg"),
                    Name("arg.bar", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

    def test_nested_function(self, parse):
        # NOTE Does not test function following, `arg` always named such
        _ast = parse(
            """
            def a_func(arg):
                def inner(arg):
                    return arg.foo
                return inner(arg)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        inner_symbol = Func("inner", ["arg"], None, None)
        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": {
                    Call("inner()", ["arg"], {}, inner_symbol),
                },
                "dels": set(),
                "gets": {
                    Name("arg"),
                    Name("arg.foo", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

    # def test_comprehensions(self, parse):
    #     _ast = parse("""
    #         def list_comp(arg):
    #             return [a.prop for a in arg.iter]
    #     """)
    #     results = FileAnalyser(_ast, RootContext(_ast)).analyse()

    #     expected = {
    #         Func("list_comp", ["arg"], None, None): {
    #             "calls": set(),
    #             "dels": set(),
    #             "gets": {
    #                 # ...
    #             },
    #             "sets": set()
    #         },
    #     }

    #     assert results == expected

    def test_getattr(self, parse):
        # Simple
        _ast = parse(
            """
            def a_func(arg):
                return getattr(arg, "attr")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

        # Nested
        _ast = parse(
            """
            def a_func(arg):
                return getattr(getattr(arg, "inner"), "outer")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("arg.inner.outer", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

        # Nested
        _ast = parse(
            """
            def a_func(arg):
                return getattr(getattr(arg.b[0], "inner"), "outer")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("arg.b[].inner.outer", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

    def test_hasattr(self, parse):
        # Simple
        _ast = parse(
            """
            def a_func(arg):
                return hasattr(arg, "attr")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

        # Nested
        _ast = parse(
            """
            def a_func(arg):
                return hasattr(hasattr(arg, "inner"), "outer")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("arg.inner.outer", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

        # Complex
        _ast = parse(
            """
            def a_func(arg):
                return hasattr(hasattr(arg.b[0], "inner"), "outer")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("arg.b[].inner.outer", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

    def test_setattr(self, parse):
        # Simple
        _ast = parse(
            """
            def a_func(arg):
                setattr(arg, "attr", "value")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": set(),
                "sets": {
                    Name("arg.attr", "arg"),
                },
            },
        }

        assert results == expected

        # Complex
        _ast = parse(
            """
            def a_func(arg):
                return setattr(arg.b[0], "attr", "value")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": set(),
                "sets": {
                    Name("arg.b[].attr", "arg"),
                },
            },
        }

        assert results == expected

    def test_delattr(self, parse):
        # Simple
        _ast = parse(
            """
            def a_func(arg):
                delattr(arg, "attr")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "gets": set(),
                "sets": set(),
                "dels": {
                    Name("arg.attr", "arg"),
                },
                "calls": set(),
            },
        }

        assert results == expected

        # Complex
        _ast = parse(
            """
            def a_func(arg):
                return delattr(arg.b[0], "attr")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "gets": set(),
                "sets": set(),
                "dels": {
                    Name("arg.b[].attr", "arg"),
                },
                "calls": set(),
            },
        }

        assert results == expected

    def test_format(self, parse, constant):
        as_func = Builtin("format", has_affect=False)

        # Simple
        _ast = parse(
            """
            def a_func(arg):
                return format(arg, "b")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": {
                    Call("format()", ["arg", constant("Str")], {}, target=as_func)
                },
                "dels": set(),
                "gets": {Name("arg")},
                "sets": set(),
            },
        }

        assert results == expected

        # Complex
        _ast = parse(
            """
            def a_func(arg):
                return format(getattr(arg, "attr"), "b")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": {
                    Call("format()", ["arg.attr", constant("Str")], {}, target=as_func)
                },
                "dels": set(),
                "gets": {
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

    def test_lambda(self, parse, capfd):
        # The Good
        _ast = parse(
            """
            global_lamb = lambda x: x.attr

            def func_one(arg):
                return global_lamb(arg)

            def func_two(arg):
                return map(lambda x: x*x, [1, 2, 3])
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        global_lamb = Func("global_lamb", ["x"], None, None)
        func_one = Func("func_one", ["arg"], None, None)
        func_two = Func("func_two", ["arg"], None, None)
        _map = Builtin("map", has_affect=False)
        expected = {
            global_lamb: {
                "gets": {
                    Name("x.attr", "x"),
                },
                "sets": set(),
                "dels": set(),
                "calls": set(),
            },
            func_one: {
                "gets": {Name("arg")},
                "sets": set(),
                "dels": set(),
                "calls": {Call("global_lamb()", ["arg"], {}, target=global_lamb)},
            },
            func_two: {
                "gets": {
                    Name("x"),
                },
                "sets": set(),
                "dels": set(),
                "calls": {
                    Call("map()", ["@Lambda", "@List"], {}, target=_map),
                },
            },
        }

        print(results)
        assert results == expected

    def test_class_init(self, parse, capfd):
        _ast = parse(
            """
            class ClassName:
                def __init__(self, arg):
                    self.attr = arg

            def a_func(blarg):
                thing = ClassName(blarg)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        cls = Class("ClassName", ["self", "arg"], None, None)
        a_func = Func("a_func", ["blarg"], None, None)
        expected = {
            cls: {
                "sets": {
                    Name("self.attr", "self"),
                },
                "gets": {Name("arg")},
                "dels": set(),
                "calls": set(),
            },
            a_func: {
                "sets": {Name("thing")},
                "gets": {Name("blarg")},
                "dels": set(),
                "calls": {Call("ClassName()", ["thing", "blarg"], {}, target=cls)},
            },
        }

        assert results == expected

    def test_return_value(self, parse, constant):
        # No return value
        _ast = parse(
            """
            def a_func(blarg):
                return
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("a_func", ["blarg"], None, None)
        expected = {
            a_func: {
                "sets": set(),
                "gets": set(),
                "dels": set(),
                "calls": set(),
            },
        }

        assert results == expected

        # Literal
        _ast = parse(
            """
            def a_func(blarg):
                return 4
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("a_func", ["blarg"], None, None)
        expected = {
            a_func: {
                "sets": set(),
                "gets": set(),
                "dels": set(),
                "calls": set(),
            },
        }

        assert results == expected

        # Local var
        _ast = parse(
            """
            def a_func(blarg):
                return blarg.attr
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("a_func", ["blarg"], None, None)
        expected = {
            a_func: {
                "sets": set(),
                "gets": {
                    Name("blarg.attr", "blarg"),
                },
                "dels": set(),
                "calls": set(),
            },
        }

        assert results == expected

        # Tuple (w/o class)
        _ast = parse(
            """
            def a_func(blarg):
                return blarg.attr, 1, a_call(blarg.another_attr)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("a_func", ["blarg"], None, None)
        expected = {
            a_func: {
                "sets": set(),
                "gets": {
                    Name("blarg.attr", "blarg"),
                    Name("blarg.another_attr", "blarg"),
                },
                "dels": set(),
                "calls": {Call("a_call()", ["blarg.another_attr"], {}, target=None)},
            },
        }

        assert results == expected

        # Class
        _ast = parse(
            """
            class MyEnum(Enum):
                first = "one"
                second = "two"

            def a_func(blarg):
                return MyEnum("one")
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("a_func", ["blarg"], None, None)
        MyEnum = Class("MyEnum", ["self", "_id"], None, None)
        expected = {
            a_func: {
                "sets": set(),
                "gets": set(),
                "dels": set(),
                "calls": {
                    Call(
                        "MyEnum()", ["@ReturnValue", constant("Str")], {}, target=MyEnum
                    ),
                },
            },
            MyEnum: {
                "sets": set(),
                "gets": {
                    Name("MyEnum.first", "MyEnum"),
                    Name("MyEnum.second", "MyEnum"),
                },
                "calls": set(),
                "dels": set(),
            },
        }

        assert results == expected

        # Tuple (w/ class)
        _ast = parse(
            """
            class MyEnum(Enum):
                first = "one"
                second = "two"

            def a_func(blarg):
                return 1, MyEnum("one"), blarg.attr
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("a_func", ["blarg"], None, None)
        MyEnum = Class("MyEnum", ["self", "_id"], None, None)
        expected = {
            a_func: {
                "sets": set(),
                "gets": {
                    Name("blarg.attr", "blarg"),
                },
                "dels": set(),
                "calls": {
                    Call(
                        "MyEnum()", ["@ReturnValue", constant("Str")], {}, target=MyEnum
                    ),
                },
            },
            MyEnum: {
                "sets": set(),
                "gets": {
                    Name("MyEnum.first", "MyEnum"),
                    Name("MyEnum.second", "MyEnum"),
                },
                "calls": set(),
                "dels": set(),
            },
        }

        print(results)
        assert results == expected

    @pytest.mark.py_3_8_plus()
    def test_walrus(self, parse):
        # General walrus use
        _ast = parse(
            """
            def fn(arg):
                if (a := arg.attr):
                    return a
                else:
                    return (b := arg.another_attr)
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("fn", ["arg"], None, None)
        expected = {
            a_func: {
                "sets": {
                    Name("a"),
                    Name("b"),
                },
                "gets": {
                    Name("a"),
                    Name("arg.attr", "arg"),
                    Name("arg.another_attr", "arg"),
                },
                "dels": set(),
                "calls": set(),
            },
        }

        print(results)
        assert results == expected

        # Basic
        _ast = parse(
            """
            def fn(arg):
                thing = (a, b := arg.attr)
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("fn", ["arg"], None, None)
        expected = {
            a_func: {
                "sets": {
                    Name("b"),
                    Name("thing"),
                },
                "gets": {
                    Name("a"),
                    Name("arg.attr", "arg"),
                },
                "dels": set(),
                "calls": set(),
            },
        }

        print(results)
        assert results == expected

        # Multiple assignment
        _ast = parse(
            """
            def fn(arg):
                x, y = (a, b := c)
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("fn", ["arg"], None, None)
        expected = {
            a_func: {
                "sets": {
                    Name("x"),
                    Name("y"),
                    Name("b"),
                },
                "gets": {
                    Name("a"),
                    Name("c"),
                },
                "dels": set(),
                "calls": set(),
            },
        }

        print(results)
        assert results == expected

    @pytest.mark.py_3_8_plus()
    def test_walrus_lambda(self, parse, capfd):
        # Walrus'd Lambda
        _ast = parse(
            """
            def fn(arg):
                other = (name := lambda *a, **k: a.attr)
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        output, _ = capfd.readouterr()
        assert "unable to unbind lambdas defined in functions" in output

        a_func = Func("fn", ["arg"], None, None)
        expected = {
            a_func: {
                "sets": {
                    Name("other"),
                    Name("name"),
                },
                "gets": {
                    Name("a.attr", "a"),
                },
                "dels": set(),
                "calls": set(),
            },
        }

        print(results)
        assert results == expected

        # Walrus multi assign w/ lambda
        _ast = parse(
            """
            def fn(arg):
                other = (alpha, beta := lambda *a, **k: a.attr)
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("fn", ["arg"], None, None)
        expected = {
            a_func: {
                "sets": {
                    Name("other"),
                    Name("beta"),
                },
                "gets": {
                    Name("alpha"),
                    Name("a.attr", "a"),
                },
                "dels": set(),
                "calls": set(),
            },
        }

        print(results)
        assert results == expected

        # Walrus multi assign w/ lambda as list
        _ast = parse(
            """
            def fn(arg):
                other = [alpha, beta := lambda *a, **k: a.attr]
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("fn", ["arg"], None, None)
        expected = {
            a_func: {
                "sets": {
                    Name("other"),
                    Name("beta"),
                },
                "gets": {
                    Name("alpha"),
                    Name("a.attr", "a"),
                },
                "dels": set(),
                "calls": set(),
            },
        }

        print(results)
        assert results == expected



"""CustomFunctionAnalyser for Python's builtin functions."""

import ast

from rattr.analyser.base import CustomFunctionAnalyser
from rattr.analyser.context import Context
from rattr.analyser.types import FuncOrAsyncFunc, FunctionIR
from rattr.analyser.util import get_dynamic_name, get_fullname


class GetattrAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "getattr"

    @property
    def qualified_name(self) -> str:
        return "getattr"

    def on_def(self, name: str, node: FuncOrAsyncFunc, ctx: Context) -> FunctionIR:
        return super().on_def(name, node, ctx)

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIR:
        return {
            "sets": set(),
            "gets": {
                get_dynamic_name(name, node, "{first}.{second}"),
            },
            "dels": set(),
            "calls": set(),
        }


class SetattrAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "setattr"

    @property
    def qualified_name(self) -> str:
        return "setattr"

    def on_def(self, name: str, node: FuncOrAsyncFunc, ctx: Context) -> FunctionIR:
        return super().on_def(name, node, ctx)

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIR:
        return {
            "sets": {
                get_dynamic_name(name, node, "{first}.{second}"),
            },
            "gets": set(),
            "dels": set(),
            "calls": set(),
        }


class HasattrAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "hasattr"

    @property
    def qualified_name(self) -> str:
        return "hasattr"

    def on_def(self, name: str, node: FuncOrAsyncFunc, ctx: Context) -> FunctionIR:
        return super().on_def(name, node, ctx)

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIR:
        return {
            "sets": set(),
            "gets": {
                get_dynamic_name(name, node, "{first}.{second}"),
            },
            "dels": set(),
            "calls": set(),
        }


class DelattrAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "delattr"

    @property
    def qualified_name(self) -> str:
        return "delattr"

    def on_def(self, name: str, node: FuncOrAsyncFunc, ctx: Context) -> FunctionIR:
        return super().on_def(name, node, ctx)

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIR:
        return {
            "sets": set(),
            "gets": set(),
            "dels": {
                get_dynamic_name(name, node, "{first}.{second}"),
            },
            "calls": set(),
        }


class SortedAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "sorted"

    @property
    def qualified_name(self) -> str:
        return "sorted"

    def on_def(self, name: str, node: FuncOrAsyncFunc, ctx: Context) -> FunctionIR:
        return super().on_def(name, node, ctx)

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIR:
        # HACK Avoid circular import
        from rattr.analyser.context.symbol import Name
        from rattr.analyser.function import FunctionAnalyser
        from rattr.analyser.ir_dag import partially_unbind

        if len(node.args) == 0:
            return super().on_call(name, node, ctx)

        fn_analyser = FunctionAnalyser(ast.FunctionDef(), ctx)

        # The first arg may be a get/call within the context
        fn_analyser.visit(node.args[0])

        # If there is a key, it should be visited
        key = None

        for kw in node.keywords:
            if kw.arg == "key":
                key = kw

        if key is None:
            return fn_analyser.func_ir

        if isinstance(key.value, ast.Lambda):
            # NOTE The lambda in `sorted` expects exactly one argument
            if len(key.value.args.args) != 1:
                raise SyntaxError

            # HACK
            #   Visit the lambda in a different FunctionAnalyser s.t. the
            #   key-argument-name can be unbound without interfering with the
            #   results from visiting the sorting target

            iterator = key.value.args.args[0].arg
            iterable = get_fullname(node.args[0], safe=True)

            lambda_ctx = Context(ctx)
            lambda_ctx.add(Name(iterator), is_argument=True)

            lambda_analyser = FunctionAnalyser(ast.FunctionDef(), lambda_ctx)
            lambda_analyser.visit(key.value.body)

            # Substitute
            lambda_analyser.func_ir = partially_unbind(
                lambda_analyser.func_ir, {iterator: iterable}
            )

            # Union
            fn_analyser.func_ir["gets"] = fn_analyser.func_ir["gets"].union(
                lambda_analyser.func_ir["gets"]
            )
            fn_analyser.func_ir["sets"] = fn_analyser.func_ir["sets"].union(
                lambda_analyser.func_ir["sets"]
            )
            fn_analyser.func_ir["dels"] = fn_analyser.func_ir["dels"].union(
                lambda_analyser.func_ir["dels"]
            )
            fn_analyser.func_ir["calls"] = fn_analyser.func_ir["calls"].union(
                lambda_analyser.func_ir["calls"]
            )
        else:
            fn_analyser.visit(key.value)

        return fn_analyser.func_ir



@rattr_results(sets={"a.attr"}, gets={"b.attr"})
def fn_a(a, b):
    return "i dont do any of that!"


@rattr_results(calls=[("fn_a", (["c", "c"], {}))])
def fn_b(c):
    print(c.some_attr)
    return fn_a(c, c)



from rattr.analyser.context.context import Context  # noqa: F401
from rattr.analyser.context.context import (  # noqa: F401
    RootContext,
    new_context,
)
from rattr.analyser.context.symbol import (  # noqa: F401
    Builtin,
    Call,
    Class,
    Func,
    Import,
    Name,
    Symbol,
)
from rattr.analyser.context.symbol_table import SymbolTable  # noqa: F401



# """Full "end-to-end" Rattr tests."""

# import pytest

# from rattr.__main__ import main, validate_arguments


# class TestRattr:

#     def test_validate_arguments(self, snippet, capfd):
#         # Non-existent file
#         with pytest.raises(FileNotFoundError):
#             validate_arguments({
#                 "<file>": snippet("no_such_file.py")
#             })

#         # Non-Python file
#         validate_arguments({"<file>": snippet("not_a_py.file")})
#         output, _ = capfd.readouterr()
#         assert "expects '.py'" in output

#         # Default filter string
#         filter_string = validate_arguments({
#             "<file>": snippet("rattr_full_one.py")
#         }).get("filter_string")
#         assert filter_string == ".+"

#     def test_rattr_full_one(self, snippet, capfd):
#         main({
#             "follow_imports": False,
#             "verbose": False,
#             "print_ast": False,
#             "debug": False,
#             "file": snippet("rattr_full_one.py"),
#             "filter_string": ".*",
#         })

#         output, _ = capfd.readouterr()

#         expected = """{
#     "a_normal_function": {
#         "sets": [],
#         "gets": [
#             "arg_one.attr"
#         ],
#         "dels": [],
#         "calls": [
#             "ValueError()"
#         ]
#     },
#     "update_attribute": {
#         "sets": [
#             "obj.target"
#         ],
#         "gets": [
#             "obj.name",
#             "new_value"
#         ],
#         "dels": [],
#         "calls": [
#             "config.get()",
#             "print()"
#         ]
#     },
#     "area": {
#         "sets": [],
#         "gets": [
#             "pi",
#             "circle.radius"
#         ],
#         "dels": [],
#         "calls": []
#     },
#     "is_a_foorbar": {
#         "sets": [
#             "filepath"
#         ],
#         "gets": [
#             "datatype.target_file",
#             "filepath",
#             "datatype.is_foo",
#             "datatype.is_bar"
#         ],
#         "dels": [],
#         "calls": [
#             "path.isfile()"
#         ]
#     },
#     "make_an_assignment": {
#         "sets": [
#             "assignee.attr"
#         ],
#         "gets": [
#             "value"
#         ],
#         "dels": [],
#         "calls": []
#     }
# }"""

#         assert "ArgumentAssignmentAssertor: line 38:" in output
#         assert expected in output

#     def test_rattr_results_one(self, snippet, capfd):
#         main({
#             "follow_imports": False,
#             "verbose": False,
#             "print_ast": False,
#             "debug": False,
#             "file": snippet("rattr_results_one.py"),
#             "filter_string": ".*",
#         })

#         output, _ = capfd.readouterr()

#         expected = """{
#     "fn_a": {
#         "sets": [
#             "a.attr"
#         ],
#         "gets": [
#             "b.attr"
#         ],
#         "dels": [],
#         "calls": []
#     },
#     "fn_b": {
#         "sets": [
#             "c.attr"
#         ],
#         "gets": [
#             "c.attr"
#         ],
#         "dels": [],
#         "calls": [
#             "fn_a()"
#         ]
#     }
# }
# """

#         assert "ExpectedDecoratorAssertor: line 2:" in output
#         assert expected in output

#     def test_rattr_results_two(self, snippet, capfd):
#         main({
#             "follow_imports": False,
#             "verbose": False,
#             "print_ast": False,
#             "debug": False,
#             "file": snippet("rattr_results_two.py"),
#             "filter_string": ".*",
#         })

#         output, _ = capfd.readouterr()

#         expected = """{
#     "fn_a": {
#         "sets": [
#             "a.attr"
#         ],
#         "gets": [
#             "b.attr"
#         ],
#         "dels": [],
#         "calls": []
#     },
#     "fn_b": {
#         "sets": [
#             "c.attr"
#         ],
#         "gets": [
#             "c.some_attr",
#             "c",
#             "c.attr"
#         ],
#         "dels": [],
#         "calls": [
#             "print()",
#             "fn_a()"
#         ]
#     }
# }
# """

#         assert "ExpectedDecoratorAssertor: line 2:" in output
#         assert expected in output



from rattr.analyser.context.symbol import (
    Call,
    Class,
    Func,
    Import,
    Name,
    Symbol,
    find_spec,
    get_module_name_and_spec,
    get_possible_module_names,
    parse_call,
    parse_name,
)


class TestSymbol:
    def test_eq(self):
        assert Name("a") == Name("a")
        assert Name("a") == Name("a", "a")
        assert Name("a") != "a"
        assert Name("a") != Name("b")
        assert Name("a") != Import("a")

    def test_hash(self):
        assert hash(Name("a")) == hash(Name("a"))
        assert hash(Name("a")) == hash(Name("a", "a"))
        assert hash(Name("a")) != hash(Name("b"))
        assert hash(Name("a")) != hash(Import("a"))

    def test_is(self):
        symbols = [
            Name("a"),
            Import("a"),
            Func("a()", [], None, None),
            Call("a()", [], {}),
            Class("a", None, None, None),
        ]
        symbol_types = [
            [Import, Func, Call, Class],  # noqa
            [Name, Func, Call, Class],  # noqa
            [Name, Import, Call, Class],  # noqa
            [Name, Import, Func, Class],  # noqa
            [
                Name,
                Import,
                Func,
                Call,
            ],  # noqa
        ]

        for symbol in symbols:
            assert symbol._is(Symbol)

        for s, _types in zip(symbols, symbol_types):
            for _type in _types:
                assert not s._is(_type)

        assert symbols[0]._is(Name)
        assert symbols[1]._is(Import)
        assert symbols[2]._is(Func)
        assert symbols[3]._is(Call)
        assert symbols[4]._is(Class)

    def test_name(self):
        assert Name("a") == Name("a", "a")

        a = Name("a")
        assert a.name == "a"
        assert a.basename == "a"

        b_dot_attr = Name("b.attr", "b")
        assert b_dot_attr.name == "b.attr"
        assert b_dot_attr.basename == "b"

    def test_import(self):
        assert Import("a") == Import("a", "a")

        # Unresolved import
        a = Import("a")
        assert a.name == "a"
        assert a.qualified_name == "a"
        assert a.module_name is None
        assert a.module_spec is None

        # Unresolved from import
        b_dot_func = Import("func", "b.func")
        assert b_dot_func.name == "func"
        assert b_dot_func.qualified_name == "b.func"
        assert b_dot_func.module_name is None
        assert b_dot_func.module_spec is None

        # Resolved stdlib import
        stdlib = Import("math")
        assert stdlib.name == "math"
        assert stdlib.qualified_name == "math"
        assert stdlib.module_name == "math"
        assert stdlib.module_spec == find_spec("math")

        # Resolved stdlib from import
        stdlib = Import("pi", "math.pi")
        assert stdlib.name == "pi"
        assert stdlib.qualified_name == "math.pi"
        assert stdlib.module_name == "math"
        assert stdlib.module_spec == find_spec("math")

        # Resolved, dotted stdlib from import
        stdlib = Import("join", "os.path.join")
        assert stdlib.name == "join"
        assert stdlib.qualified_name == "os.path.join"
        assert stdlib.module_name == "os.path"
        assert stdlib.module_spec == find_spec("os.path")

    def test_func(self):
        # def fn()
        fn = Func("fn", [], None, None)
        assert fn.name == "fn"
        assert fn.args == []
        assert fn.vararg is None
        assert fn.kwarg is None
        assert not fn.is_async

        # async def async_fn(a, b='val', *c, **d)
        async_fn = Func("async_fn", ["a", "b"], "c", "d", is_async=True)
        assert async_fn.name == "async_fn"
        assert async_fn.args == ["a", "b"]
        assert async_fn.vararg == "c"
        assert async_fn.kwarg == "d"
        assert async_fn.is_async

    def test_call(self):
        # fn()
        call = Call("fn", [], {})
        assert call.name == "fn"
        assert call.args == []
        assert call.kwargs == {}

        # fn(a, b, c=d)
        call = Call("fn", ["a", "b"], {"c": "d"})
        assert call.name == "fn"
        assert call.args == ["a", "b"]
        assert call.kwargs == {"c": "d"}

    def test_class(self):
        # class ClassName
        # Always initialised with None's (outside of tests, for convenience)
        # None's imply that initialiser has not been seen
        cls = Class("ClassName", None, None, None)
        assert cls.name == "ClassName"
        assert cls.args is None
        assert cls.vararg is None
        assert cls.kwarg is None


class TestSymbolUtils:
    def test_parse_name(self):
        assert parse_name("var") == Name("var")
        assert parse_name("var.attr") == Name("var.attr", "var")

        assert parse_name("*var") == Name("*var", "var")
        assert parse_name("*var.attr") == Name("*var.attr", "var")

        complex_name = "*var.mth().res[].attr"
        assert parse_name(complex_name) == Name(complex_name, "var")

    def test_parse_call(self):
        pos, named = [], {}
        assert parse_call("func()", (pos, named)) == Call("func()", pos, named)

        pos, named = ["a", "b", "c"], {"d": "e", "f": "g"}
        assert parse_call("func()", (pos, named)) == Call("func()", pos, named)

        # Allow omit brackets
        pos, named = [], {}
        assert parse_call("func", (pos, named)) == Call("func()", pos, named)

    def test_get_possible_module_names(self):
        expected = ["a.b.c.d", "a.b.c", "a.b", "a"]
        assert get_possible_module_names("a.b.c.d") == expected

        expected = ["os.path.join", "os.path", "os"]
        assert get_possible_module_names("os.path.join") == expected

        expected = [""]
        assert get_possible_module_names("") == expected

        expected = ["a_name_without_dots"]
        assert get_possible_module_names("a_name_without_dots") == expected

    def test_get_module_name_and_spec(self):
        assert get_module_name_and_spec("i.am.not.a.module") == (None, None)

        expected = ("math", find_spec("math"))
        assert get_module_name_and_spec("math") == expected
        assert get_module_name_and_spec("math.pi") == expected

        expected = ("os", find_spec("os"))
        assert get_module_name_and_spec("os") == expected

        expected = ("os.path", find_spec("os.path"))
        assert get_module_name_and_spec("os.path") == expected

        assert get_module_name_and_spec("os") != get_module_name_and_spec("os.path")

        m = "rattr.analyser"
        expected = (m, find_spec(m))
        assert get_module_name_and_spec(m) == expected

        m = "rattr.analyser.util"
        expected = (m, find_spec(m))
        assert get_module_name_and_spec(f"{m}.get_module_spec") == expected



"""Represent the IR of functions as a DAG, for simplification."""

from copy import deepcopy
from dataclasses import dataclass
from typing import Dict, List, Optional, Set, Tuple, Union

from rattr import config, error
from rattr.analyser.context import (
    Builtin,
    Call,
    Class,
    Func,
    Import,
    Name,
    Symbol,
)
from rattr.analyser.types import FileIR, FunctionIR, ImportsIR
from rattr.analyser.util import (
    is_blacklisted_module,
    is_excluded_name,
    is_pip_module,
    is_stdlib_module,
    module_name_from_file_path,
    remove_call_brackets,
)


def __prefix(func: Func) -> str:
    """HACK We no longer have `culprit` so manually construct prefix."""
    if func is None:
        return ""

    if func.defined_in is not None and config.show_path:
        file = error.format_path(func.defined_in)
    else:
        file = ""

    return "\033[1m{}:\033[0m".format(file)


def __resolve_target_and_ir(
    callee: Call,
    file_ir: FileIR,
    imports_ir: ImportsIR,
) -> Tuple[Func, FunctionIR]:
    """Helper function for `resolve_function` and `resolve_class`."""
    if callee.target in file_ir:
        return callee.target, file_ir[callee.target]

    filename = callee.target.defined_in
    module = module_name_from_file_path(filename)

    if module is None:
        raise ModuleNotFoundError(f"unable to find module for '{filename}'")

    module_ir = imports_ir.get(module)

    if module_ir is None:
        raise ImportError

    if callee.target not in module_ir:
        raise ImportError

    return callee.target, module_ir[callee.target]


def resolve_function(
    callee: Call,
    file_ir: FileIR,
    imports_ir: ImportsIR,
    caller: Optional[Func] = None,
) -> Union[Tuple[None, None], Tuple[Func, FunctionIR]]:
    _msg = f"{__prefix(caller)} unable to resolve call to " f"'{callee.target.name}'"

    if caller is not None:
        _msg = f"{_msg} in '{caller.name}'"

    try:
        func, ir = __resolve_target_and_ir(callee, file_ir, imports_ir)
    except ImportError:
        if is_excluded_name(callee.target.name):
            error.error(f"{_msg}, the call target matches an exclusion")
        elif "." not in callee.name:
            error.error(f"{_msg}, likely a nested function or ignored")
        else:
            error.info(f"{_msg}'")
        return None, None

    return func, ir


def resolve_class(
    callee: Call,
    file_ir: FileIR,
    imports_ir: ImportsIR,
    caller: Optional[Func] = None,
) -> Union[Tuple[None, None], Tuple[Func, FunctionIR]]:
    _where = __prefix(caller)

    try:
        cls, ir = __resolve_target_and_ir(callee, file_ir, imports_ir)
    except ImportError:
        error.error(f"{_where} unable to resolve initialiser for '{callee.name}'")
        return None, None

    return cls, ir


def resolve_import(
    name: str,
    target: Import,
    imports_ir: ImportsIR,
    caller: Optional[Func] = None,
) -> Union[Tuple[None, None], Tuple[Func, FunctionIR]]:
    """Return the `Func` and IR for the given import."""
    _where = __prefix(caller)

    module = target.module_name
    module_ir: FileIR = imports_ir.get(module, None)

    as_name, qualified_name = target.name, target.qualified_name

    if is_blacklisted_module(module):
        return None, None

    if not config.follow_imports:
        error.info(f"{_where} ignoring call to imported function '{target.name}'")
        return None, None

    if not config.follow_pip_imports and is_pip_module(module):
        error.info(
            f"{_where} ignoring call to function '{target.name}' imported "
            f"from pip installed module '{module}'"
        )
        return None, None

    if not config.follow_stdlib_imports and is_stdlib_module(module):
        error.info(
            f"{_where} ignoring call to function '{target.name}' imported "
            f"from stdlib module '{module}'"
        )
        return None, None

    if module is None:
        raise ValueError(f"Target {target.name} has no module name")

    if module_ir is None:
        raise ImportError(f"Import '{module}' not found")

    local_name = name.replace(as_name, qualified_name).replace(f"{module}.", "")
    local_name = remove_call_brackets(local_name)

    new_target: Optional[Symbol] = module_ir.context.get(local_name)
    if isinstance(new_target, (Func, Class)):
        ir = module_ir.get(new_target)

        # NOTE If the imported function is ignored then it will have no IR
        if ir is None:
            error.error(
                f"{_where} unable to resolve imported callable '{local_name}'"
                f" in '{module}', it is likely ignored"
            )
            return None, None

        return new_target, ir

    if isinstance(new_target, Import):
        return resolve_import(new_target.name, new_target, imports_ir, caller)

    # NOTE
    #   When reaching here the target may be a call to a method on an imported
    #   instance

    if new_target is None and "." in local_name:
        error.info(
            f"{__prefix(caller)} unable to resolve call to method "
            f"'{local_name}' in import '{module}'"
        )
        return None, None

    error.error(
        f"{__prefix(caller)} unable to resolve call to '{local_name}' in "
        f"import '{module}'"
    )
    return None, None


def get_callee_target(
    callee: Call,
    file_ir: FileIR,
    imports_ir: ImportsIR,
    caller: Optional[Func] = None,
) -> Union[Tuple[None, None], Tuple[Func, FunctionIR]]:
    """Return the `Symbol` and IR of the called function."""
    # TODO Should this trigger a warning?
    if callee.target is None:
        return None, None

    if isinstance(callee.target, Builtin):
        return None, None

    if isinstance(callee.target, Func):
        return resolve_function(callee, file_ir, imports_ir, caller)

    # NOTE Procedural parameter, etc, can't be resolved
    if isinstance(callee.target, Name):
        return None, None

    if isinstance(callee.target, Class):
        return resolve_class(callee, file_ir, imports_ir, caller)

    if isinstance(callee.target, Import):
        return resolve_import(callee.name, callee.target, imports_ir, caller)

    raise TypeError("'callee' must be a CalleeTarget")


def partially_unbind_name(symbol: Name, new_basename: str) -> Name:
    """Return a new symbol bound to the new base name."""
    if symbol.basename == new_basename:
        return Name(symbol.name, symbol.basename)

    new_name = symbol.name
    if symbol.name.startswith("*"):
        new_name = symbol.name.replace(f"*{symbol.basename}", f"*{new_basename}", 1)
    else:
        new_name = symbol.name.replace(symbol.basename, new_basename, 1)

    return Name(new_name, new_basename)


def partially_unbind(func_ir: FunctionIR, swaps: Dict[str, str]) -> FunctionIR:
    """Return the partially unbound results for the given function."""
    return {
        "sets": {
            partially_unbind_name(n, swaps.get(n.basename, n.basename))
            for n in func_ir["sets"]
        },
        "gets": {
            partially_unbind_name(n, swaps.get(n.basename, n.basename))
            for n in func_ir["gets"]
        },
        "dels": {
            partially_unbind_name(n, swaps.get(n.basename, n.basename))
            for n in func_ir["dels"]
        },
        "calls": func_ir["calls"],
    }


def construct_swap(func: Func, call: Call) -> Dict[str, str]:
    """Return the map of local function names to bound names."""
    swaps: Dict[str, str] = dict()

    f_args = deepcopy(func.args)

    c_args = deepcopy(call.args)
    c_kwargs = deepcopy(call.kwargs)

    # Python call must be ([pos_arg | *'d], ..., [named_arg | **'d], ...)
    # Python function def must be (pos_arg, ..., named_arg, ..., *'d?, **'d?)

    for target, replacement in zip(f_args, c_args):
        swaps[target] = replacement
    f_args = f_args[len(c_args) :]

    # Ensure no ambiguities
    for target in swaps.keys():
        if target not in c_kwargs.keys():
            continue

        error.fatal(f"{__prefix(func)} '{target}' given by position and name")

        return swaps

    # Allow no match as a default may be present
    for target in f_args:
        if target not in c_kwargs:
            continue
        swaps[target] = c_kwargs[target]
        del c_kwargs[target]

    if len(c_kwargs) > 0:
        error.error(f"{__prefix(func)} unexpected named arguments {c_kwargs}")

    return swaps


@dataclass
class IrDagNode:
    """Represent a function call for simplification."""

    call: Call
    func: Func
    func_ir: FunctionIR
    file_ir: FileIR
    imports_ir: ImportsIR

    def __post_init__(self) -> None:
        self.children: List[IrDagNode] = list()

    def populate(self, seen: Optional[Set[Call]] = None) -> Set[Call]:
        """Populate this node, and it's children, recursively.

        In principle, BFS the calls to construct the DAG. Calls may have cycles
        (i.e. direct or indirect recursion), however these can be ignored if
        the same function is called with the same arguments -- thus eliminating
        cycles; producing a DAG.

        """
        if seen is None:
            seen = set()

        # Find children
        for callee in self.func_ir["calls"]:
            if callee in seen:
                continue

            _foc, _foc_ir = get_callee_target(
                callee, self.file_ir, self.imports_ir, self.func
            )

            # NOTE Could not find func/init (undef'd, <str>.join(), etc)
            if _foc is None or _foc_ir is None:
                continue

            self.children.append(
                IrDagNode(callee, _foc, _foc_ir, self.file_ir, self.imports_ir)
            )

            seen.add(callee)

        # Populate children
        for child in self.children:
            seen = child.populate(seen)

        return seen

    def simplify(self) -> FunctionIR:
        """Return the IR modified to include dependent calls.

        NOTE
            Assumes that the root IrDagNode has been populated via `populate`!

        """
        # Leafs are already simplified
        if len(self.children) == 0:
            return deepcopy(self.func_ir)

        # Simplified non-terminal nodes are the combination of themselves
        # and their children partially unbound
        child_irs = [(c.simplify(), c) for c in self.children]

        simplified: FunctionIR = deepcopy(self.func_ir)

        for child_ir, child in child_irs:
            swaps = construct_swap(child.func, child.call)
            unbound_child = partially_unbind(child_ir, swaps)

            simplified["sets"] = simplified["sets"].union(unbound_child["sets"])

            simplified["gets"] = simplified["gets"].union(unbound_child["gets"])

            simplified["dels"] = simplified["dels"].union(unbound_child["dels"])

        return simplified



"""Rattr analyser context and utilities.

A context holds a symbol table for a given scope (file, function, class, etc)
within a given file. All contexts for a file for a tree with unidirectional
references from child context to parent context.

The root context of a file contains, amongst other things, the Python builtin
functions such as `print`, etc.

"""

import ast
from contextlib import contextmanager
from dataclasses import replace as copy_dataclass
from typing import Any, Generator, Iterable, Optional, Set, TypeVar, Union

from rattr import config, error
from rattr.analyser.context.symbol import (
    Builtin,
    CallTarget,
    Class,
    Func,
    Import,
    Name,
    Symbol,
    get_module_name_and_spec,
    get_possible_module_names,
)
from rattr.analyser.context.symbol_table import SymbolTable
from rattr.analyser.types import AnyAssign, Constant, Literal, ast_NamedExpr
from rattr.analyser.util import (
    PYTHON_BUILTINS,
    Changes,
    assignment_is_one_to_one,
    enter_file,
    get_absolute_module_name,
    get_assignment_targets,
    get_contained_walruses,
    get_fullname,
    get_function_def_args,
    get_starred_imports,
    has_affect,
    is_blacklisted_module,
    is_method_on_primitive,
    is_relative_import,
    is_starred_import,
    lambda_in_rhs,
    module_name_from_file_path,
    remove_call_brackets,
    unravel_names,
)

_Context = TypeVar("_Context", bound="Context")


@contextmanager
def new_context(container: Any) -> Generator:
    """Push then pop a new context."""
    if not hasattr(container, "context"):
        raise AttributeError("'container' needs the attribute 'context'")

    parent = container.context
    container.context = Context(parent)
    yield
    container.context = parent


class Context:
    """Hold the context for some scope.

    A context and it's parent, parent's parent, etc form a tree of contexts
    that correlates with the AST of the related file.

    The parent and the parent's parent, etc of a context, C, are collectively
    refferred to as the "ancestors" of C.

    """

    def __init__(self, parent: Optional[_Context]) -> None:
        self.parent = parent
        self.symbol_table = SymbolTable()
        self.file = config.current_file

    def add(self, symbol: Symbol, is_argument: bool = False) -> None:
        """Add the given symbol to the context.

        If the symbol is already present in the context, or an ancestor, then
        it will not be re-added.

        If the symbol `is_argument`, then the symbol will always be added
        regardless of whether-or-not it is declared in an ancestor.

        """
        if not is_argument and symbol in self:
            return

        self.symbol_table.add(symbol)

    def add_all(self, symbols: Iterable[Symbol]) -> None:
        """Add all of the given symbols (assumes none are arguments)."""
        for s in symbols:
            self.add(s)

    def remove(self, name: str) -> None:
        """Remove the symbol matching the given name from the context."""
        if self.declares(name):
            self.symbol_table.remove(name)
            return

        if self.parent is None:
            return

        self.parent.remove(name)

    def remove_all(self, names: Iterable[str]) -> None:
        """Remove all of the given symbols."""
        for n in names:
            self.remove(n)

    def get(self, name: str) -> Optional[Symbol]:
        """Return the symbol matching `name`, in this context or an ancestor.

        If `name` exists in some context, C, and an ancestor of C, then the
        most constrained (closet) definition is used -- i.e. the definition in
        C.

        """
        if self.parent is None:
            return self.symbol_table.get(name)

        return self.symbol_table.get(name) or self.parent.get(name)

    def get_call_target(
        self, callee: str, culprit: ast.Call, warn: bool = True
    ) -> Optional[Symbol]:
        """Return the target of the given AST call, if within the context."""
        if not isinstance(callee, str) or not callee.endswith("()"):
            raise ValueError("'callee' must a string ending with '()'")

        name = remove_call_brackets(callee).replace("*", "")
        stripped = name.replace("*", "").replace("[]", "").replace("()", "")

        if stripped.startswith("@"):
            return None

        target = self.get(name)

        # NOTE
        #   Method on some LHS in context, but not a call to an implicitly
        #   imported function
        _lhs_target = self.get(stripped.split(".")[0])
        if target is None and not isinstance(_lhs_target, Import):
            target = _lhs_target

        # NOTE
        #   If target is not defined locally (or explicitly imported locally),
        #   it could be implicitly imported.
        #       `from math import pi`   ->  explicitly imported as `pi`
        #       `import math`           ->  implicitly imported as `math.pi`
        if target is None:
            modules = map(self.get, get_possible_module_names(stripped))
            defined = list(filter(lambda m: m is not None, modules))
            _lhs_target = next(iter(defined), None)

            if isinstance(_lhs_target, Import):
                local_name = name.replace(f"{_lhs_target.name}.", "")
                qualified_name = f"{_lhs_target.qualified_name}.{local_name}"
                target = Import(name, qualified_name)
            else:
                target = _lhs_target

        # NOTE
        #   If declared in this scope and called, it is likely an argument or a
        #   lambda (handled elsewhere).
        #   If it is an argument, then it is either a procedural parameter or a
        #   call to a method on an argument; distinguished by "." being in the
        #   callee name.
        if warn and target is not None and not isinstance(target, CallTarget.__args__):
            if "." not in callee and self.declares(target.name):
                error.error(
                    f"unable to resolve call to '{name}', likely a procedural "
                    f"parameter",
                    culprit,
                )
            elif "." in callee:
                # TODO Once method support is added, elevate this to an error
                error.info(f"unable to resolve call to method '{name}'", culprit)
            else:
                error.error(f"'{name}' is not callable", culprit)

        if warn and target is None and not is_method_on_primitive(name):
            error.error(f"unable to resolve call to '{name}'", culprit)

        if warn and target is not None and callee.endswith("()()"):
            error.warning("unable to resolve call result of call", culprit)

        return target

    def declares(self, name: str) -> bool:
        """Return `True` if `name` was declared in this context."""
        return name in self.symbol_table.names()

    def __contains__(self, symbol: Union[str, Symbol]) -> bool:
        """Return `True` if `symbol` was declared in this or an ancestor."""
        name = symbol if isinstance(symbol, str) else symbol.name

        if self.parent is None:
            return self.declares(name)

        return self.declares(name) or (name in self.parent)

    def get_root(self) -> _Context:
        """Return the root context for the related file."""
        if self.parent is None:
            return self

        return self.parent.get_root()

    def is_import(self, name: str) -> bool:
        """Return `True` if `name` refferres to an import in this context."""
        return isinstance(self.get(name), Import)

    # ----------------------------------------------------------------------- #
    # Registration helpers
    # ----------------------------------------------------------------------- #

    def add_identifiers_to_context(self, assignment: ast.expr) -> None:
        for name in unravel_names(assignment):
            self.add(Name(name))

    def del_identifiers_from_context(self, assignment: ast.expr):
        for name in unravel_names(assignment):
            self.remove(name)

    def push_arguments_to_context(self, arguments: ast.arguments) -> None:
        all_arguments = [*arguments.args, arguments.vararg, arguments.kwarg]

        for arg in [a for a in all_arguments if a]:
            self.add(Name(arg.arg), is_argument=True)

    # ----------------------------------------------------------------------- #
    # Syntactic sugar
    # ----------------------------------------------------------------------- #

    def expand_starred_imports(self) -> _Context:
        """Expand starred imports to normal imports."""
        seen: Set[str] = set()

        # Create initial starred list for BFS
        starred = get_starred_imports(self.symbol_table.symbols(), seen)

        # BFS the unseen * imports
        for _i in starred:
            if _i.module_spec is None or _i.module_spec.origin is None:
                error.error(
                    f"unable to resolve import '{_i.name}' while expanding "
                    f"syntactic sugar 'from {_i.qualified_name} import *' "
                    f"in {self.file}"
                )
                continue

            if _i.module_spec.origin in seen:
                continue

            with open(_i.module_spec.origin, "r") as f:
                _i_ast = ast.parse(f.read())

            with enter_file(_i.module_spec.origin):
                _i_ctx = RootContext(_i_ast)

            seen.add(_i.module_spec.origin)

            starred += get_starred_imports(_i_ctx.symbol_table.symbols(), seen)

            # Tread `from x import *` as syntactic sugar for
            # `from x import a, b, c, ...`
            for s in _i_ctx.symbol_table.symbols():
                self.add(Import(s.name, f"{_i.qualified_name}.{s.name}"))

        return self


def _new_import_symbol(
    name: str,
    qualified_name: str,
    module: str,
    node: ast.AST,
) -> Import:
    """Return the Import symbol for the given name; warn if spec not found."""
    if not isinstance(node, (ast.Import, ast.ImportFrom)):
        raise TypeError

    _import = Import(name, qualified_name)

    if is_blacklisted_module(module):
        return _import

    if _import.module_spec is None:
        error.fatal(f"unable to find module '{module}'", node)

    return _import


class RootContext(Context):
    """Return the root context for the given module.

    Supports:
        * Import, ImportFrom
        * Assign, AugAssign
        * FunctionDef, AsyncFunctionDef
        * ClassDef
        * If
        * For, AsyncFor
        * While
        * Try
        * With, AsyncWith

    """

    module_level_attributes = [
        "__annotations__",
        "__builtins__",
        "__cached__",
        "__doc__",
        "__file__",
        "__loader__",
        "__name__",
        "__package__",
        "__spec__",
    ]

    def __new__(self, module: ast.Module) -> Any:  # Context, make mypy happy
        if not isinstance(module, ast.Module):
            raise TypeError("The root context can only exist for a module")

        context = Context(None)

        # Register Python module level attributes
        for attribute in RootContext.module_level_attributes:
            context.add(Name(attribute))

        # Register the Python builtins
        for name in PYTHON_BUILTINS:
            context.add(Builtin(name, has_affect=has_affect(name)))

        # Register the top-level symbols
        for node in module.body:
            RootContext.register(context, node)

        return context

    @staticmethod
    def register(context: Context, node: ast.AST) -> None:
        """Register the given node."""
        name = node.__class__.__name__
        register_method = getattr(RootContext, f"register_{name}", None)

        if not register_method:
            error.fatal("unsupported top-level statement", node)
            return

        return register_method(context, node)

    # ----------------------------------------------------------------------- #
    # Registration, self is of type Context @see __new__() and register()
    # ----------------------------------------------------------------------- #

    def register_Import(self, node: ast.Import) -> None:
        """Register the given imports with qualified names.

        >>> import math
        Import(name="math", qualified_name="math")

        >>> import math as m
        Import(name="m", qualified_name="math")

        >>> import math, more_math
        Import(name="math", qualified_name="math")
        Import(name="more_math", qualified_name="more_math")

        """
        if len(node.names) > 1:
            error.info("do not import multiple modules on oneline", node)

        for module in node.names:
            _import = _new_import_symbol(
                module.asname or module.name, module.name, module.name, node
            )
            self.add(_import)

    def register_starred_import(self, node: ast.ImportFrom) -> None:
        """Helper method for starred imports."""
        if self.file is None or not self.file.endswith("__init__.py"):
            error.warning(
                f"do not use 'from {node.module} import *', be explicit", node
            )

        module = node.module

        # Allow starred imports to be relative
        if node.level > 0:
            base = module_name_from_file_path(config.current_file)
            module = get_absolute_module_name(base, node.level, node.module)

            if get_module_name_and_spec(module) == (None, None):
                error.error("unable to resolve relative starred import", node)

        self.add(_new_import_symbol("*", module, module, node))

    def register_relative_import(self, node: ast.ImportFrom) -> None:
        """Helper method for relative imports."""
        base = module_name_from_file_path(config.current_file)

        if base is None:
            error.fatal("unable to resolve parent in relative import", node)

        module = get_absolute_module_name(base, node.level, node.module)

        if get_module_name_and_spec(module) == (None, None):
            error.error("unable to resolve relative import", node)

        for target in node.names:
            _import = _new_import_symbol(
                target.asname or target.name, f"{module}.{target.name}", module, node
            )
            self.add(_import)

    def register_named_imports(self, node: ast.ImportFrom) -> None:
        """Helper method for named imports, i.e. `from M import name, ...`."""
        for target in node.names:
            _import = _new_import_symbol(
                target.asname or target.name,
                f"{node.module}.{target.name}",
                node.module,
                node,
            )
            self.add(_import)

    def register_ImportFrom(self, node: ast.ImportFrom) -> None:
        """Register the given from-imports with qualified names.

        >>> from math import pi
        Import(name="pi", qualified_name="math.pi")

        >>> from math import pi as pie
        Import(name="pie", qualified_name="math.pi")

        >>> from math import *
        Import(name="*", qualified_name="math")

        """
        if is_starred_import(node):
            RootContext.register_starred_import(self, node)
            return

        if is_relative_import(node):
            RootContext.register_relative_import(self, node)
            return

        RootContext.register_named_imports(self, node)

    def register_AnyAssign(self, node: AnyAssign) -> None:
        targets = get_assignment_targets(node)

        if lambda_in_rhs(node):
            if not assignment_is_one_to_one(node):
                error.fatal("lambda assignment must be one-to-one", node)
            else:
                name = get_fullname(targets[0])
                self.add(Func(name, *get_function_def_args(node.value)))
            return

        # Walrus operator in the right-hand side of containing assignment
        for walrus in get_contained_walruses(node):
            with Changes(self.symbol_table, keys_fn=lambda t: t.values()) as diff:
                RootContext.register_NamedExpr(self, walrus)

            # If we have "a = (b := lambda ...)" then "a" will have been registered as a
            # Name when it just be a Func with the same signature as the just-added "b"
            if lambda_in_rhs(walrus):
                if len(diff.added) == 1 and node.value == walrus:
                    lhs = get_assignment_targets(node)[0]
                    self.add(
                        copy_dataclass(list(diff.added)[0], name=get_fullname(lhs))
                    )
                elif len(diff.added) > 1:
                    raise NotImplementedError("Multiple deeply nested walruses")

        for target in targets:
            self.add_identifiers_to_context(target)

    def register_Assign(self, node: ast.Assign) -> None:
        RootContext.register_AnyAssign(self, node)

    def register_AnnAssign(self, node: ast.AnnAssign) -> None:
        RootContext.register_AnyAssign(self, node)

    def register_AugAssign(self, node: ast.AugAssign) -> None:
        RootContext.register_AnyAssign(self, node)

    def register_Delete(self, node: ast.Delete) -> None:
        for target in node.targets:
            self.del_identifiers_from_context(target)

        error.warning(
            "avoid 'del' at module-level, Rattr will assume the target has "
            "been removed before reaching any child context (functions, "
            "classes, etc)",
            node,
        )

    def register_NamedExpr(self, node: ast_NamedExpr) -> None:
        RootContext.register_AnyAssign(self, node)

    # ----------------------------------------------------------------------- #
    # Functions, classes, etc
    # ----------------------------------------------------------------------- #

    def register_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.add(Func(node.name, *get_function_def_args(node)))

    def register_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self.add(Func(node.name, *get_function_def_args(node), is_async=True))

    def register_ClassDef(self, node: ast.ClassDef) -> None:
        # NOTE
        #   Create as constructor/initialiser-less class, populate those symbol
        #   attributes when analysing the class itself
        self.add(Class(node.name, None, None, None))

    # ----------------------------------------------------------------------- #
    # Top-level blocks of statements
    # ----------------------------------------------------------------------- #

    def register_stmts(self, *stmts) -> None:
        for stmt in stmts:
            RootContext.register(self, stmt)

    def register_If(self, node: ast.If) -> None:
        RootContext.register_stmts(self, *node.body, *node.orelse)

    def register_For(self, node: ast.For) -> None:
        RootContext.register_stmts(self, *node.body, *node.orelse)

    def register_AsyncFor(self, node: ast.For) -> None:
        RootContext.register_stmts(self, *node.body, *node.orelse)

    def register_While(self, node: ast.While) -> None:
        RootContext.register_stmts(self, *node.body, *node.orelse)

    def register_Try(self, node: ast.Try) -> None:
        bodies = node.body + node.orelse + node.finalbody
        handlers = [s for h in node.handlers for s in h.body]

        RootContext.register_stmts(self, *bodies, *handlers)

    def register_With(self, node: ast.With) -> None:
        RootContext.register_stmts(self, *node.body)

    def register_AsyncWith(self, node: ast.AsyncWith) -> None:
        RootContext.register_stmts(self, *node.body)

    # ----------------------------------------------------------------------- #
    # Explicit ignores: docstrings, etc.
    # ----------------------------------------------------------------------- #

    def register_Expr(self, node: ast.Expr) -> None:
        """Ignore standalone expressions.

        Literal | Constant
            Just a value (absent from an assignment, thus ignoreable),
            e.g. a docstring.

        Call
            Likely calling an entry-point or part of configuration -- shouldn't
            affect context and thus can be ignored.

        Lambda
            If lambda is not an assignment RHS or a call argument (i.e. just
            "floating"), then disallow it.

        """
        if isinstance(node.value, Literal.__args__):
            return

        if isinstance(node.value, Constant.__args__):
            return

        if isinstance(node.value, ast.Call):
            return

        if isinstance(node.value, ast.Lambda):
            return error.fatal("top-level lambdas must be named", node)

        _name = node.__class__.__name__
        error.warning(f"unexpected top-level expression 'ast.{_name}'", node)
